import org.kohsuke.args4j.CmdLineParser;

import java.io.*;

public class ConvertCode {
    public final static int CHANGE_TYPE_LAST_CONSTANT = 0;
    public final static int CHANGE_TYPE_INSTANT = 1;
    public final static int CHANGE_TYPE_EXPONENTIAL = 2;
    public final static int GENERATIONS_INDEX = 3;
    public final static int N_INDEX = 5;
    public final static int CHANGTYPE_INDEX = 7;
    protected double startTime = 0;

    protected int numOfIntervals;
    protected double N0;
    protected  int sampleSize;

    protected double[] timeIntervals;
    protected double[] ratiosOfNiOverN0;
    protected int[] changeTypes;

    protected double[] timeIntervalsIn2N0;
    private double[] ratiosOfNiOverN0_log;
    private double[] ratiosOfNiOverN0_inverse;

    private double[] changeNs;

    protected void setParameter_generationTimes(double[] generationTimes, double[] sizeOfNs, int[] changeTypes){
        numOfIntervals = generationTimes.length;
        if( (numOfIntervals==sizeOfNs.length) && (numOfIntervals==changeTypes.length) ){
            N0 = sizeOfNs[0];
            this.changeTypes = changeTypes;
            timeIntervalsIn2N0 = new double[numOfIntervals];
            ratiosOfNiOverN0 = new double[numOfIntervals];
            ratiosOfNiOverN0_log = new double[numOfIntervals];
            ratiosOfNiOverN0_inverse = new double[numOfIntervals];

            double oneOverN0 = 1.0 / N0;

            for (int i = 0; i < numOfIntervals; i ++) {
                timeIntervalsIn2N0[i] = generationTimes[i] * 0.5 * oneOverN0;
                ratiosOfNiOverN0[i] = sizeOfNs[i] * oneOverN0;
                ratiosOfNiOverN0_log[i]     = Math.log(ratiosOfNiOverN0[i]);
                ratiosOfNiOverN0_inverse[i] = 1.0d / ratiosOfNiOverN0[i];
            }

            timeIntervals = setTimeIntervals(timeIntervalsIn2N0, ratiosOfNiOverN0, ratiosOfNiOverN0_log, ratiosOfNiOverN0_inverse, changeTypes);
            changeNs = setChangeNs_baseOnTimeIntervalsIn2N0(timeIntervalsIn2N0, ratiosOfNiOverN0, ratiosOfNiOverN0_log, changeTypes);

        } else{
            System.out.println("Error! Please check numOfIntervals!");
        }
    }

    private double[] setChangeNs_baseOnTimeIntervalsIn2N0(double[] timeIntervalsIn2N0, double[] ratiosOfNiOverN0, double[] ratiosOfNiOverN0_log, int[] changeTypes) {
        double[] changeNs = new double[numOfIntervals];
        for (int i = 0; i < numOfIntervals; i ++) {
            int changeType = changeTypes[i];
            if (changeType == CHANGE_TYPE_EXPONENTIAL) {
                changeNs[i] = (ratiosOfNiOverN0_log[i+1] - ratiosOfNiOverN0_log[i]) / (timeIntervalsIn2N0[i+1] - timeIntervalsIn2N0[i]);
            }
        }
        return changeNs;
    }

    private double[] setTimeIntervals(double[] timeIntervalsIn2N0, double[] ratiosOfNiOverN0, double[] ratiosOfNiOverN0_log, double[] ratiosOfNiOverN0_inverse, int[] changeTypes) {
        double[] timeIntervals = new double[numOfIntervals];
        if (numOfIntervals > 0) {
            timeIntervals[0] = 0;
            for(int i = 1; i < numOfIntervals; i ++) {
                int changeType = changeTypes[i - 1];
                switch (changeType) {
                    case CHANGE_TYPE_LAST_CONSTANT: // The last interval
                        // timeIntervalsIn2N0[i] = calculateTimeIn2N0(timeIntervals[i]);
                        break;
                    case CHANGE_TYPE_INSTANT:
//				timeIntervalsIn2N0[i] = ( timeIntervals[i] - timeIntervals[i-1] ) * ratiosOfNiOverN0[i-1] + timeIntervalsIn2N0[i-1];
                        timeIntervals[i] = ( timeIntervalsIn2N0[i] - timeIntervalsIn2N0[i-1] ) * ratiosOfNiOverN0_inverse[i-1] + timeIntervals[i-1];
                        break;
                    case CHANGE_TYPE_EXPONENTIAL://
                        double diff_logRatio = ratiosOfNiOverN0_log[i] - ratiosOfNiOverN0_log[i-1];
//					double diff_inverseRatio = ratiosOfNiOverN0_inverse[i-1] - ratiosOfNiOverN0_inverse[i];
                        if (Math.abs(diff_logRatio) < 0.0001) {
//						timeIntervalsIn2N0[i] = ( timeIntervals[i] - timeIntervals[i-1] ) * ratiosOfNiOverN0[i-1] + timeIntervalsIn2N0[i-1];
                            timeIntervals[i] = ( timeIntervalsIn2N0[i] - timeIntervalsIn2N0[i-1] ) * ratiosOfNiOverN0_inverse[i-1] + timeIntervals[i-1];
                        }
                        else {
//						timeIntervalsIn2N0[i] =(timeIntervals[i] - timeIntervals[i-1]) / diff_inverseRatio * diff_logRatio + timeIntervalsIn2N0[i-1];
                            timeIntervals[i] = ( timeIntervalsIn2N0[i] - timeIntervalsIn2N0[i-1] ) / diff_logRatio * ( ratiosOfNiOverN0_inverse[i-1] - ratiosOfNiOverN0_inverse[i] ) + timeIntervals[i-1];
                        }
                        break;
                    default:
                        // Do nothing.
                }
            }
        }
        return timeIntervals;
    }

    protected String strChangeType(int changeType){
        String str = "";
        if(changeType == CHANGE_TYPE_LAST_CONSTANT) {
            str = "last";
        }
        else if(changeType == CHANGE_TYPE_EXPONENTIAL ) {
            str = "exponential";
        }
        else if(changeType == CHANGE_TYPE_INSTANT){
            str = "instant";
        }
        else {
            String calssName = Thread.currentThread().getStackTrace()[1].getClassName();
            String methodName = Thread.currentThread().getStackTrace()[1].getMethodName();
            str = "[ERROR: " + calssName + "." + methodName+ "](changeType)";

        }
        return str;
    }

    protected String covertMsCode(){
        String s = "ms demographic parameters: ";
        for(int i = 0; i < numOfIntervals; i ++){
            double time = timeIntervalsIn2N0[i] * 0.5;
            s += " -eN " + time;
            s += " " + ratiosOfNiOverN0[i];
            if(changeTypes[i] == CHANGE_TYPE_EXPONENTIAL){
                s += " -eG " + time;
                double alpha = - 2 *  (ratiosOfNiOverN0_log[i] - ratiosOfNiOverN0_log[i+1]) / (timeIntervalsIn2N0[i] - timeIntervalsIn2N0[i+1]);
                s += " " + alpha;
            }
            else{
                s += " -eG " + time;
                s += " 0.0";
            }
        }
        return s;
    }

    protected void run(String inputFile) throws IOException {
        FileReader fileReader = new FileReader(inputFile);
        LineNumberReader lineNumberReader = new LineNumberReader(fileReader);
        String line;
        String[] elements;
        while ((line = lineNumberReader.readLine()) != null) {
            if (line.startsWith("******** Best: ********")) {
                line  = lineNumberReader.readLine();
                elements = line.trim().split("\\s+");
                int numOfIntervals = Integer.parseInt(elements[elements.length - 1]);
                double[] generations = new double[numOfIntervals];
                double[] sizeOfNs = new double[numOfIntervals];
                int [] changeTypes = new int[numOfIntervals];
                for (int i = 0; i < numOfIntervals; i++) {
                    elements = lineNumberReader.readLine().trim().split("\\s+");
                    generations[i] = Double.parseDouble(elements[GENERATIONS_INDEX]);
                    sizeOfNs[i] = Double.parseDouble(elements[N_INDEX].replaceAll(",", ""));
                    changeTypes[i] = Integer.parseInt(elements[CHANGTYPE_INDEX]);
                }
                setParameter_generationTimes(generations, sizeOfNs, changeTypes);

                System.out.println(covertMsCode());
            }
        }


    }

    public static void main(String[] args) throws Exception {
        ConvertCodeInput convertCodeInput = new ConvertCodeInput();
        CmdLineParser parser = new CmdLineParser(convertCodeInput);
        parser.parseArgument(args);

        ConvertCode convertCode = new ConvertCode();
        String inputFile = convertCodeInput.getInputFile();
        convertCode.run(inputFile);
    }
}
