import cern.jet.random.engine.MersenneTwister64;
import cern.jet.random.Poisson;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.Calendar;
import java.util.List;

public class EstimatorForSinglePopModel {
	protected double mutRatePerGenome;
	protected boolean forbidInstantaneousGrowth;
	protected boolean forbidExponentialGrowth;
	protected boolean forbidInstantaneousDecline;
	protected boolean forbidExponentialDecline;
	protected int fixedNumOfIntervals;
	protected double likelihoodRate;
	protected int numOfOmitStartSfs;
	protected int numOfOmitEndSfs;
	protected int numOfCollapsedEndSfs;
	protected int numOfRepeats;
	protected double timeUpperBound;
	protected double gammaTimeLUS;
	protected double gammaPopSizeLUS;
	protected double minChangeOfNPerPhase;
	protected double maxChangeOfNPerPhase;
	protected double minRatioOfNOverN0;
	protected double maxRatioOfNOverN0;
    
	protected boolean useFoldedSFS = false; // Haipeng

    public int sampleSize;
    public int[] obsSFS;

    public MersenneTwister64 random;
    public Poisson poisson;

    public TableForLinearModel table = new TableForLinearModel();

    protected DecimalFormat intFormat  = new DecimalFormat("###,###,###");

    public EstimatorForSinglePopModel() {
        int randSeed = Math.abs((int)(Calendar.getInstance().getTime().getTime()));
        random = new MersenneTwister64(randSeed);
        poisson = new Poisson(0, random);
    }

    public EstimatorForSinglePopModel(int randSeed) {
        random = new MersenneTwister64(randSeed);
        poisson = new Poisson(0, random);
    }

    public void setObsSFS(int[] obsSFS) {
        this.obsSFS = obsSFS.clone();
        this.sampleSize = obsSFS.length + 1;
    }

    public PopSizeModelExponentialLinearInstant runOptimizeFreeTimes(List<int[]> obsSFSs, List<Integer> sampleSizes) throws IOException {

        TableForLinearModel[] tables = new TableForLinearModel[sampleSizes.size()];
        for (int i = 0; i < tables.length; i++) {
            tables[i] = new TableForLinearModel();
            tables[i].setSampleSize(sampleSizes.get(i));
            tables[i].initialMatrix();
        }
        MinFitnessOfNs optimizer = new MinFitnessOfNs();
        optimizer.setNumOfOmitStartSfs(numOfOmitStartSfs);
        optimizer.setNumOfOmitEndSfs(numOfOmitEndSfs);
        optimizer.setNumOfCollapsedEndSfs(numOfCollapsedEndSfs);
        
        optimizer.setMutRatePerGenome(mutRatePerGenome);
        optimizer.setRandom(random);
        optimizer.setForbidInstantaneousGrowth(forbidInstantaneousGrowth);
        optimizer.setForbidInstantaneousDecline(forbidInstantaneousDecline);
        optimizer.setForbidExponentialGrowth(forbidExponentialGrowth);
        optimizer.setForbidExponentialDecline(forbidExponentialDecline);
        optimizer.setTimeUpperBound(timeUpperBound);
        optimizer.setGammaTimeLUS(gammaTimeLUS);
        optimizer.setGammaPopSizeLUS(gammaPopSizeLUS);
        optimizer.setMinChangeOfNPerPhase(minChangeOfNPerPhase);
        optimizer.setMaxChangeOfNPerPhase(maxChangeOfNPerPhase);
        optimizer.setMinRatioOfNOverN0(minRatioOfNOverN0);
        optimizer.setMaxRatioOfNOverN0(maxRatioOfNOverN0);
        
        optimizer.setUseFoldedSFS(useFoldedSFS); // Haipeng

        double ratioOfOmitStart = (double) numOfOmitStartSfs / sampleSizes.get(0);
        double ratioOfOmitLast = (double) numOfOmitEndSfs / sampleSizes.get(0);
        int usedBoundary = obsSFSs.size();
        for (int i = 0; i < obsSFSs.size(); i++) {
            int omitStart = (int) Math.round(ratioOfOmitStart * sampleSizes.get(i));
            int omitLast = (int) Math.round(ratioOfOmitLast * sampleSizes.get(i));
            if (obsSFSs.get(i).length - omitStart - omitLast <= 0) {
                usedBoundary = i;
                break;
            }
            int[] usedSFS = new int[obsSFSs.get(i).length - omitStart - omitLast];
            for (int j = 0; j < usedSFS.length; j++) {
                usedSFS[j] = obsSFSs.get(i)[j + omitStart];
            }
        }

        Result4SinglePopHistory bestResult = null;

        if (fixedNumOfIntervals == 0){
            int numOfIntervals = 1;
            while (true){
                Result4SinglePopHistory bestResultGivenNumOfParas = optimizer.optimize(numOfIntervals, tables, obsSFSs, usedBoundary);
                for (int repeatIndex = 1; repeatIndex < numOfRepeats; repeatIndex ++) {
                    Result4SinglePopHistory oneRepeat = optimizer.optimize(numOfIntervals, tables, obsSFSs, usedBoundary);
                    if (oneRepeat.fitness < bestResultGivenNumOfParas.fitness) {
                        bestResultGivenNumOfParas = oneRepeat;
                    }
                }

                System.out.println("numOfIntervals = " + numOfIntervals);
                printResult(bestResultGivenNumOfParas);

                if (bestResult == null) {
                    bestResult = bestResultGivenNumOfParas;
                } else {
                    if (bestResult.fitness < bestResultGivenNumOfParas.fitness * (likelihoodRate /100 + 1)) {
                        break;
                    } else {
                        bestResult = bestResultGivenNumOfParas;
                    }
                }
                numOfIntervals++;
            }
        } else {
            bestResult = optimizer.optimize(fixedNumOfIntervals, tables, obsSFSs, usedBoundary);
            for (int repeatIndex = 1; repeatIndex < numOfRepeats; repeatIndex ++) {
                Result4SinglePopHistory oneRepeat = optimizer.optimize(fixedNumOfIntervals, tables, obsSFSs, usedBoundary);
                if (oneRepeat.fitness < bestResult.fitness) {
                    bestResult = oneRepeat;
                }
            }
        }
        System.out.println("******** Best: ********");
        System.out.println("numOfIntervals = " + bestResult.popSizeChangeTypes.length);

        printResult(bestResult);

        PopSizeModelExponentialLinearInstant popModel = new PopSizeModelExponentialLinearInstant();
        popModel.setParameters(bestResult.startTimeOfIntervals, bestResult.ratiosOfNiOverN0, bestResult.popSizeChangeTypes);
        popModel.setN0(bestResult.N0);

        return popModel;
    }

    public PopSizeModelExponentialLinearInstant runOptimizeFreeTimes(int[] obsSFS) throws IOException {
        setObsSFS(obsSFS);

        double obsTreeLength = 0;
        double[] obsBranchLengthsInGen = new double[obsSFS.length];
        int obsK = 0;
        for (int i = 0; i < obsSFS.length; i ++) {
            obsBranchLengthsInGen[i] = obsSFS[i] / mutRatePerGenome;
            obsK += obsSFS[i];
            obsTreeLength += obsBranchLengthsInGen[i];
        }


        table.setSampleSize(sampleSize);
        table.initialMatrix();

        MinFitnessOfNs optimizer = new MinFitnessOfNs();
        optimizer.setNumOfOmitStartSfs(numOfOmitStartSfs);
        optimizer.setNumOfOmitEndSfs(numOfOmitEndSfs);
        optimizer.setMutRatePerGenome(mutRatePerGenome);
        optimizer.setRandom(random);
        optimizer.setForbidInstantaneousGrowth(forbidInstantaneousGrowth);
        optimizer.setForbidInstantaneousDecline(forbidInstantaneousDecline);
        optimizer.setForbidExponentialGrowth(forbidExponentialGrowth);
        optimizer.setForbidExponentialDecline(forbidExponentialDecline);
        optimizer.setTimeUpperBound(timeUpperBound);
        optimizer.setGammaTimeLUS(gammaTimeLUS);
        optimizer.setGammaPopSizeLUS(gammaPopSizeLUS);
        optimizer.setMinChangeOfNPerPhase(minChangeOfNPerPhase);
        optimizer.setMaxChangeOfNPerPhase(maxChangeOfNPerPhase);
        optimizer.setMinRatioOfNOverN0(minRatioOfNOverN0);
        optimizer.setMaxRatioOfNOverN0(maxRatioOfNOverN0);
        
        optimizer.setUseFoldedSFS(useFoldedSFS); // Haipeng

        optimizer.setObsSFS(obsSFS);

        int[] usedSFS = new int[obsSFS.length - numOfOmitStartSfs - numOfOmitEndSfs];
        for (int i = 0; i < usedSFS.length; i++) {
            usedSFS[i] = obsSFS[i + numOfOmitStartSfs];
        }

        Result4SinglePopHistory bestResult = null;

        int maxNumOfIntervalsSearched;
        if (fixedNumOfIntervals == 0){
            maxNumOfIntervalsSearched = sampleSize;
            for (int numOfIntervals = 1; numOfIntervals <= maxNumOfIntervalsSearched; numOfIntervals++) {
                Result4SinglePopHistory bestResultGivenNumOfParas = optimizer.optimize(numOfIntervals, table);
                for (int repeatIndex = 1; repeatIndex < numOfRepeats; repeatIndex ++) {
                    Result4SinglePopHistory oneRepeat = optimizer.optimize(numOfIntervals, table);
                    if (oneRepeat.fitness < bestResultGivenNumOfParas.fitness) {
                        bestResultGivenNumOfParas = oneRepeat;
                    }
                }
                System.out.println("numOfIntervals = " + numOfIntervals);
                printResult(bestResultGivenNumOfParas);

                if (bestResult == null) {
                    bestResult = bestResultGivenNumOfParas;
                } else {
                    if (bestResult.fitness < bestResultGivenNumOfParas.fitness * (likelihoodRate / 100 + 1)) {
                        break;
                    } else {
                        bestResult = bestResultGivenNumOfParas;
                    }
                }
            }
        } else {
            bestResult = optimizer.optimize(fixedNumOfIntervals, table);
            for (int repeatIndex = 1; repeatIndex < numOfRepeats; repeatIndex ++) {
                Result4SinglePopHistory oneRepeat = optimizer.optimize(fixedNumOfIntervals, table);
                if (oneRepeat.fitness < bestResult.fitness) {
                    bestResult = oneRepeat;
                }
            }
        }
        System.out.println("******** Best: ********");
        System.out.println("numOfIntervals = " + bestResult.popSizeChangeTypes.length);

        printResult(bestResult);

        PopSizeModelExponentialLinearInstant popModel = new PopSizeModelExponentialLinearInstant();
        popModel.setParameters(bestResult.startTimeOfIntervals, bestResult.ratiosOfNiOverN0, bestResult.popSizeChangeTypes);
        popModel.setN0(bestResult.N0);

        return popModel;
    }

    protected void printResult(Result4SinglePopHistory result){
        double N0 = result.N0;

        double[] ratiosOfNiOverN0 = result.ratiosOfNiOverN0;
        double[] packedStartTimeOfIntervals = result.startTimeOfIntervals;
        int[] changeTypes = result.popSizeChangeTypes;

        PopSizeModelExponentialLinearInstant popModel = new PopSizeModelExponentialLinearInstant();
        popModel.setParameters(packedStartTimeOfIntervals, ratiosOfNiOverN0, changeTypes);

        for (int j = 0; j < ratiosOfNiOverN0.length; j ++) {
            System.out.println("standardTime: " + packedStartTimeOfIntervals[j] + " generations: " + String.format("%.2f", popModel.calculateTimeIn2N0(packedStartTimeOfIntervals[j]) * N0 * 2) + " N: " + intFormat.format(N0 * ratiosOfNiOverN0[j]) + " changeTypes: " + changeTypes[j]);

        }
        System.out.println("maxLogL = -" + result.fitness);

        System.out.println();
    }

    public void setMutRatePerGenome(double mutRatePerGenome) {
        this.mutRatePerGenome = mutRatePerGenome;
    }

    public void setForbidInstantaneousGrowth(boolean forbidInstantaneousGrowth) {
        this.forbidInstantaneousGrowth = forbidInstantaneousGrowth;
    }

    public void setForbidExponentialGrowth(boolean forbidExponentialGrowth) {
        this.forbidExponentialGrowth = forbidExponentialGrowth;
    }

    public void setForbidInstantaneousDecline(boolean forbidInstantaneousDecline) {
        this.forbidInstantaneousDecline = forbidInstantaneousDecline;
    }

    public void setForbidExponentialDecline(boolean forbidExponentialDecline) {
        this.forbidExponentialDecline = forbidExponentialDecline;
    }

    public void setFixedNumOfIntervals(int fixedNumOfIntervals) {
        this.fixedNumOfIntervals = fixedNumOfIntervals;
    }

    public void setLikelihoodRate(double likelihoodRate) {
        this.likelihoodRate = likelihoodRate;
    }

    public void setNumOfOmitStartSfs(int numOfOmitStartSfs) {
        this.numOfOmitStartSfs = numOfOmitStartSfs;
    }

    public void setNumOfOmitEndSfs(int numOfOmitEndSfs) {
        this.numOfOmitEndSfs = numOfOmitEndSfs;
    }
    
    public void setNumOfCollapsedEndSfs(int numOfCollapsedEndSfs) {
        this.numOfCollapsedEndSfs = numOfCollapsedEndSfs;
    }

    public void setNumOfRepeats(int numOfRepeats) {
        this.numOfRepeats = numOfRepeats;
    }

    public void setTimeUpperBound(double timeUpperBound) {
        this.timeUpperBound = timeUpperBound;
    }

    public void setGammaTimeLUS(double gammaTimeLUS) {
        this.gammaTimeLUS = gammaTimeLUS;
    }

    public void setGammaPopSizeLUS(double gammaPopSizeLUS) {
        this.gammaPopSizeLUS = gammaPopSizeLUS;
    }

    public void setMinChangeOfNPerPhase(double minChangeOfNPerPhase) {
        this.minChangeOfNPerPhase = minChangeOfNPerPhase;
    }

    public void setMaxChangeOfNPerPhase(double maxChangeOfNPerPhase) {
        this.maxChangeOfNPerPhase = maxChangeOfNPerPhase;
    }

    public void setMinRatioOfNOverN0(double minRatioOfNOverN0) {
        this.minRatioOfNOverN0 = minRatioOfNOverN0;
    }

    public void setMaxRatioOfNOverN0(double maxRatioOfNOverN0) {
        this.maxRatioOfNOverN0 = maxRatioOfNOverN0;
    }
    
    // Haipeng
    public void setUseFoldedSFS(boolean useFoldedSFS) {
        this.useFoldedSFS = useFoldedSFS;
    }
}
