public class PopSizeModelExponentialLinearInstant implements PopSizeModelInterface {
    private double N0 = 0;

    private final double startTime = 0;
    private double endTime = Double.MAX_VALUE;

    private int currentIntervalIndex = 0;

    private int numOfIntervals = 1;

    private double[] timeIntervals = new double[numOfIntervals];

    private double[] timeIntervalsIn2N0 = new double[numOfIntervals];

    private int[] changeTypes = new int[numOfIntervals];

    private double[] ratiosOfNiOverN0 = new double[numOfIntervals];
    private double[] ratiosOfNiOverN0Log = new double[numOfIntervals];
    private double[] ratiosOfNiOverN0Inverse = new double[numOfIntervals];

    private double[] changeNs = new double[numOfIntervals];
    private double[] changeNsInverse = new double[numOfIntervals];

    public PopSizeModelExponentialLinearInstant() {
        ratiosOfNiOverN0[0] = 1;
        ratiosOfNiOverN0Log[0] = 0;
        ratiosOfNiOverN0Inverse[0] = 1;

        changeTypes[0] = CHANGE_TYPE_LAST_CONSTANT;
    }

    public PopSizeModelExponentialLinearInstant clone() {
        PopSizeModelExponentialLinearInstant cloned = new PopSizeModelExponentialLinearInstant();

        double[] timeIntervals = this.timeIntervals.clone();
        // startTimes of the intervals.
        // TimeIntervals are the sum of the all the time periods in the unit of 2N_i generations.
        double[] ratiosOfNiOverN0 = this.ratiosOfNiOverN0.clone(); //backward continuous

        int[] changeTypes = this.changeTypes.clone();

        cloned.setEndTime(this.endTime);
        cloned.setParameters(timeIntervals, ratiosOfNiOverN0, changeTypes);

        return cloned;
    }

    public void setEndTime(double endTime) {
        this.endTime = endTime;
    }

    private int getIndexOfInsertValue(double[] values, double toBeInserted) {
        int length = values.length;
        if (values[length - 1] <= toBeInserted) {
            return length;
        }
        else {
            return getIndexOfInsertValue(values, toBeInserted, 0, values.length - 1);
        }
    }

    private int getIndexOfInsertValue(double[] values, double toBeInserted, int low, int high) {
        int mid = (low + high) / 2;
        if (mid == low) {
            return high;
        }
        else {
            double midValue = values[mid];
            if (toBeInserted < midValue) {
                return getIndexOfInsertValue(values, toBeInserted, low, mid);
            }
            else {
                return getIndexOfInsertValue(values, toBeInserted, mid, high);
            }
        }
    }

    private int[] getCompareTimeIndex(double[] timeIntervals, double[] tabulatedTimes){
        int[] timeIndex = new int[numOfIntervals];
        for(int i = 0; i < numOfIntervals; i++) {
            timeIndex[i] = getIndexOfInsertValue(tabulatedTimes, timeIntervals[i])-1;
        }
        return timeIndex;
    }

    public PopSizeModelInstantChange convert(double[] tabulatedTimes) {
        int numOfConvertedIntervals = 0;
        int[] compareTimeIndex = getCompareTimeIndex(timeIntervals,tabulatedTimes);
        for (int i = 0; i < numOfIntervals; i ++) {
            if(changeTypes[i] == CHANGE_TYPE_EXPONENTIAL) {
                if(timeIntervals[i+1] == tabulatedTimes[compareTimeIndex[i+1]]) {
                    numOfConvertedIntervals += compareTimeIndex[i+1] - compareTimeIndex[i];
                }
                else {
                    numOfConvertedIntervals += compareTimeIndex[i+1] - compareTimeIndex[i] + 1;
                }
            }
            else {
                numOfConvertedIntervals ++;
            }
        }

        double[] instantChanges_TimeIntervals = new double[numOfConvertedIntervals];
        double[] instantChanges_RatiosOfNiOverN0 = new double[numOfConvertedIntervals];

        int currentConvertIntervalIndex = 0;
        for (int i = 0; i < numOfIntervals; i ++) {
            int changeType = changeTypes[i];

            if (changeType == CHANGE_TYPE_EXPONENTIAL) {
                double startRatiosOfNiOverN0_inverse = ratiosOfNiOverN0Inverse[i];
                double endRatiosOfNiOverN0_inverse = ratiosOfNiOverN0Inverse[i+1];

                double startTimeInterval = timeIntervals[i];
                double endTimeInterval = timeIntervals[i+1];

                int startTimeCompareIndex = compareTimeIndex[i];
                int endTimeCompareIndex = compareTimeIndex[i+1];

                if(timeIntervals[i+1] == tabulatedTimes[endTimeCompareIndex]) {
                    endTimeCompareIndex --;
                }

                double inverseChangeRate = (endRatiosOfNiOverN0_inverse - startRatiosOfNiOverN0_inverse) / (endTimeInterval - startTimeInterval);
                double previousRatiosOfNiOverN0inverse = startRatiosOfNiOverN0_inverse;
                double crrentRatiosOfNiOverN0inverse = 0;

                instantChanges_TimeIntervals[currentConvertIntervalIndex] = startTimeInterval;
                currentConvertIntervalIndex ++;

                for (int j = startTimeCompareIndex+1; j <= endTimeCompareIndex; j ++) {
                    instantChanges_TimeIntervals[currentConvertIntervalIndex] = tabulatedTimes[j];

                    crrentRatiosOfNiOverN0inverse = startRatiosOfNiOverN0_inverse+(tabulatedTimes[j] - startTimeInterval)*inverseChangeRate;
                    instantChanges_RatiosOfNiOverN0[currentConvertIntervalIndex-1] = 2.0/(previousRatiosOfNiOverN0inverse+crrentRatiosOfNiOverN0inverse);
                    previousRatiosOfNiOverN0inverse = crrentRatiosOfNiOverN0inverse;
                    currentConvertIntervalIndex ++;
                }
                instantChanges_RatiosOfNiOverN0[currentConvertIntervalIndex-1] = 2.0/(previousRatiosOfNiOverN0inverse+endRatiosOfNiOverN0_inverse);
            }
            else {
                instantChanges_RatiosOfNiOverN0[currentConvertIntervalIndex] = ratiosOfNiOverN0[i];
                instantChanges_TimeIntervals[currentConvertIntervalIndex] = timeIntervals[i];
                currentConvertIntervalIndex ++;
            }
        }

        PopSizeModelInstantChange instantChange = new PopSizeModelInstantChange();
        instantChange.setParameters(instantChanges_TimeIntervals, instantChanges_RatiosOfNiOverN0);

        return instantChange;
    }

    @Override
    public double calculateTimeIn2N0(double standardTime) {
        int indexOfTimeInterval = getIndexOfTimeIntervalsInStepStone(standardTime);
        double temp = calculateLocalTimeIn2N0GivenIndexOfTimeInterval(indexOfTimeInterval,
                standardTime - timeIntervals[indexOfTimeInterval]);
        double timeIn2N0 = timeIntervalsIn2N0[indexOfTimeInterval] + temp;

        return timeIn2N0;
    }

    private double calculateLocalTimeIn2N0GivenIndexOfTimeInterval(int indexOfTimeInterval, double localTime) {
        double localTimeIn2N0 = Double.NaN;
        int changeType = changeTypes[indexOfTimeInterval];
        switch (changeType) {
            case CHANGE_TYPE_LAST_CONSTANT:
                localTimeIn2N0 = ratiosOfNiOverN0[indexOfTimeInterval] * localTime;
                break;
            case CHANGE_TYPE_INSTANT:
                localTimeIn2N0 = ratiosOfNiOverN0[indexOfTimeInterval] * localTime;
                break;
            case CHANGE_TYPE_EXPONENTIAL:
                if( Math.abs(changeNs[indexOfTimeInterval]) < 0.00001) {
                    localTimeIn2N0 = localTime / ratiosOfNiOverN0[indexOfTimeInterval];
                }
                else {
                    double NOverN0_log = -Math.log(ratiosOfNiOverN0Inverse[indexOfTimeInterval] - localTime * changeNs[indexOfTimeInterval]);
                    localTimeIn2N0 = (NOverN0_log - ratiosOfNiOverN0Log[indexOfTimeInterval]) * changeNsInverse[indexOfTimeInterval];
                }
                break;
            default:
        }
        return localTimeIn2N0;
    }

    private double getRatioOfNiOverN0_EXPONENTIAL_inverse(double startIntervalTime, double endIntervalTime, double startRatioOfNiOverN0_inverse, double endRatioOfNiOverN0_inverse, double time) {
        double ratio_inverse = solveLinearRatioOfNiOverN0(startIntervalTime, endIntervalTime, startRatioOfNiOverN0_inverse, endRatioOfNiOverN0_inverse, time);
        return ratio_inverse;
    }

    private double solveLinearRatioOfNiOverN0(double startTime, double endTime, double startRatioOfNiOverN0, double endRatioOfNiOverN0, double time) {
        double ratio = Double.NaN;
        if( (time > startTime) && (time < endTime) ) {
            ratio = (time - startTime) * endRatioOfNiOverN0 + (endTime - time) * startRatioOfNiOverN0;
            ratio /= (endTime - startTime);
        }
        else {
            if(time == startTime) {
                ratio = startRatioOfNiOverN0;
            }
        }
        return ratio;
    }

    @Override
    public double getRatioOfNiOverN0(double time) {
        int indexOfTimeInterval = getIndexOfTimeIntervalsInStepStone(time);
        return getRatioOfNiOverN0(indexOfTimeInterval, time);
    }

    public double getRatioOfNiOverN0(int indexOfTimeInterval, double time) {
        double ratioOfNiOverN0 = Double.NaN;
        int changeType = changeTypes[indexOfTimeInterval];
        switch (changeType) {
            case CHANGE_TYPE_LAST_CONSTANT: // The last interval
                ratioOfNiOverN0 = ratiosOfNiOverN0[indexOfTimeInterval];
                break;
            case CHANGE_TYPE_INSTANT:
                ratioOfNiOverN0 = ratiosOfNiOverN0[indexOfTimeInterval];
                break;
            case CHANGE_TYPE_EXPONENTIAL:
                ratioOfNiOverN0 = 1.0d / getRatioOfNiOverN0_EXPONENTIAL_inverse(timeIntervals[indexOfTimeInterval], timeIntervals[indexOfTimeInterval+1], ratiosOfNiOverN0Inverse[indexOfTimeInterval], ratiosOfNiOverN0Inverse[indexOfTimeInterval+1], time);
                break;
            default:
                // Do nothing.
        }

        return ratioOfNiOverN0;
    }

    @Override
    public int getNumOfIntervals() {
        return numOfIntervals;
    }

    private int getIndexOfTimeIntervals(double time, double[] timeIntervals) {
        if (time <  timeIntervals[0]) {
            return 0;
        }
        if (time >= timeIntervals[numOfIntervals - 1]) {
            return (numOfIntervals - 1);
        }

        return getIndexOfTimeIntervals(time, 0, numOfIntervals - 1, timeIntervals);
    }


    private int getIndexOfTimeIntervals(double time, int lowIndex, int highIndex, double[] timeIntervals) {
        int midIndex = (highIndex + lowIndex) / 2;

        int index = -1;
        if (lowIndex == midIndex) {
            index = lowIndex;
        }
        else {
            double midStartTime = timeIntervals[midIndex];
            if (time < midStartTime) {
                index = getIndexOfTimeIntervals(time, lowIndex, midIndex, timeIntervals);
            }
            else {
                index = getIndexOfTimeIntervals(time, midIndex, highIndex, timeIntervals);
            }
        }

        return index;
    }

    @Override
    public int getIndexOfTimeIntervalsInStepStone(double time) {
        return getIndexOfTimeIntervalsInStepStone(time, timeIntervals);

    }

    @Override
    public int getIndexOfTimeIntervalsInStepStone(double time, double[] timeIntervals) {
        if (timeIntervals[currentIntervalIndex] <= time) {
            if (currentIntervalIndex < numOfIntervals - 1) {

                if (time < timeIntervals[currentIntervalIndex + 1]) {
                    return currentIntervalIndex;
                }
                else {
                    currentIntervalIndex ++;
                    return getIndexOfTimeIntervalsInStepStone(time, timeIntervals);
                }
            }
            else {
                return currentIntervalIndex;
            }
        }
        else {

            if (timeIntervals[currentIntervalIndex - 1] <= time) {
                currentIntervalIndex --;
                return currentIntervalIndex;
            }
            else {
                currentIntervalIndex --;
                return getIndexOfTimeIntervalsInStepStone(time, timeIntervals);
            }
        }

    }

    @Override
    public int getIndexOfTimeIntervals(double time) {
        return getIndexOfTimeIntervals(time, timeIntervals);
    }

    @Override
    public void setParameters(double[] timeIntervals, double[] ratiosOfNiOverN0) {
        this.timeIntervals = timeIntervals;
        this.ratiosOfNiOverN0 = ratiosOfNiOverN0;
        this.numOfIntervals = timeIntervals.length;
        timeIntervalsIn2N0 = setTimeIntervalsIn2N0(timeIntervals, ratiosOfNiOverN0, ratiosOfNiOverN0Log, ratiosOfNiOverN0Inverse);
    }

    @Override
    public void setParameters(double[] timeIntervals, double[] ratiosOfNiOverN0, int[] changeTypes) {
        this.timeIntervals = timeIntervals;
        this.ratiosOfNiOverN0 = ratiosOfNiOverN0;
        this.changeTypes = changeTypes;

        this.numOfIntervals = timeIntervals.length;

        this.ratiosOfNiOverN0Log = new double[numOfIntervals];
        this.ratiosOfNiOverN0Inverse = new double[numOfIntervals];

        for (int i = 0; i < numOfIntervals; i ++) {
            this.ratiosOfNiOverN0Log[i]     = Math.log(ratiosOfNiOverN0[i]);
            this.ratiosOfNiOverN0Inverse[i] = 1.0d / ratiosOfNiOverN0[i];
        }

        changeNs = setChangeNs(timeIntervals, ratiosOfNiOverN0Log, ratiosOfNiOverN0Inverse, changeTypes);
        changeNsInverse = setChangeNsInverse(changeNs);
        timeIntervalsIn2N0 = setTimeIntervalsIn2N0(timeIntervals, ratiosOfNiOverN0, ratiosOfNiOverN0Log, ratiosOfNiOverN0Inverse, changeTypes);
    }

    private double[] setChangeNs(double[] timeIntervals, double[] ratiosOfNiOverN0_log, double[] ratiosOfNiOverN0_inverse, int[] changeTypes) {
        double[] changeNs = new double[numOfIntervals];
        for (int i = 0; i < numOfIntervals; i ++) {
            int changeType = changeTypes[i];

            if (changeType == CHANGE_TYPE_EXPONENTIAL) {
                changeNs[i] = (ratiosOfNiOverN0_inverse[i] - ratiosOfNiOverN0_inverse[i+1]) / (timeIntervals[i+1] - timeIntervals[i]);;
            }
        }
        return changeNs;
    }

    private double[] setChangeNsInverse(double[] changeNs) {
        int changeNsLength = changeNs.length;
        double[] changeNs_inverse = new double[changeNsLength];
        for (int i = 0; i < changeNsLength; i ++) {
            if (changeNs[i] != 0) {
                changeNs_inverse[i] = 1 / changeNs[i];
            }
        }
        return changeNs_inverse;
    }

    private double[] setTimeIntervalsIn2N0(double[] timeIntervals, double[] ratiosOfNiOverN0, double[] ratiosOfNiOverN0Log, double[] ratiosOfNiOverN0Inverse) {
        double[] timeIntervalsIn2N0 = new double[numOfIntervals];
        timeIntervalsIn2N0[0] = 0;
        for (int i = 1; i <numOfIntervals; i++) {
            timeIntervalsIn2N0[i] = ( timeIntervals[i] - timeIntervals[i-1] ) * ratiosOfNiOverN0[i-1] + timeIntervalsIn2N0[i-1];
        }
        return timeIntervalsIn2N0;
    }

    private double[] setTimeIntervalsIn2N0(double[] timeIntervals, double[] ratiosOfNiOverN0, double[] ratiosOfNiOverN0_log, double[] ratiosOfNiOverN0_inverse, int[] changeTypes) {
        double[] timeIntervalsIn2N0 = new double[numOfIntervals];
        if (numOfIntervals > 0) {
            timeIntervalsIn2N0[0] = startTime;
            for(int i = 1; i < numOfIntervals; i ++) {
                int changeType = changeTypes[i - 1];
                switch (changeType) {
                    case CHANGE_TYPE_LAST_CONSTANT:
                        break;
                    case CHANGE_TYPE_INSTANT:
                        timeIntervalsIn2N0[i] = ( timeIntervals[i] - timeIntervals[i-1] ) * ratiosOfNiOverN0[i-1] + timeIntervalsIn2N0[i-1];
                        break;
                    case CHANGE_TYPE_EXPONENTIAL:
                        double diff_inverseRatio = ratiosOfNiOverN0_inverse[i-1] - ratiosOfNiOverN0_inverse[i];
                        if (Math.abs(diff_inverseRatio) < 0.0001) {
                            timeIntervalsIn2N0[i] = ( timeIntervals[i] - timeIntervals[i-1] ) * ratiosOfNiOverN0[i-1] + timeIntervalsIn2N0[i-1];
                        }
                        else {
                            timeIntervalsIn2N0[i] =(timeIntervals[i] - timeIntervals[i-1]) / diff_inverseRatio * ( ratiosOfNiOverN0_log[i] - ratiosOfNiOverN0_log[i-1] ) + timeIntervalsIn2N0[i-1];
                        }
                        break;
                    default:
                }
            }
        }

        return timeIntervalsIn2N0;
    }

    public double getN0() {
        return N0;
    }

    public void setN0(double n0) {
        N0 = n0;
    }

    public double[] getTimeIntervals() {
        return timeIntervals;
    }

    public int[] getChangeTypes() {
        return changeTypes;
    }

}
