public class PopSizeModelInstantChange implements PopSizeModelInterface{
    public int numOfIntervals = 1;
    public double[] timeIntervals   = new double[numOfIntervals];
    public double[] ratiosOfNiOverN0 = new double[numOfIntervals];

    public int currentIntervalIndex = 0;

    public PopSizeModelInstantChange() {
        ratiosOfNiOverN0[0] = 1.0d;
    }

    @Override
    public void setParameters(double[] timeIntervals, double[] ratiosOfNiOverN0) {
        this.timeIntervals = timeIntervals;
        this.ratiosOfNiOverN0 = ratiosOfNiOverN0;

        this.numOfIntervals = timeIntervals.length;
    }

    @Override
    public int getNumOfIntervals() {
        return numOfIntervals;
    }

    @Override
    public double getRatioOfNiOverN0(double time) {
        if (time > 20) {
            throw new IllegalArgumentException("time = " + time + "! Not in the unit of 2N!");
        }

        int indexOfTimeIntervals = getIndexOfTimeIntervalsInStepStone(time);

        return ratiosOfNiOverN0[indexOfTimeIntervals];
    }

    @Override
    public int getIndexOfTimeIntervals(double time) {
        if (time <  timeIntervals[0]) {
            return 0;
        }
        if (time >= timeIntervals[numOfIntervals - 1]) {
            return (numOfIntervals - 1);
        }

        return getIndexOfTimeIntervals(time, 0, numOfIntervals - 1);
    }

    protected int getIndexOfTimeIntervals(double time, int lowIndex, int highIndex) {
        int midIndex = (highIndex + lowIndex) / 2;

        int index = -1;
        if (lowIndex == midIndex) {
            index = lowIndex;
        }
        else {
            double midStartTime = timeIntervals[midIndex];
            if (time < midStartTime) {
                index = getIndexOfTimeIntervals(time, lowIndex, midIndex);
            }
            else {
                index = getIndexOfTimeIntervals(time, midIndex, highIndex);
            }
        }

        return index;
    }

    @Override
    public double calculateTimeIn2N0(double standardTime) {
        double timeIn2N0 = 0;
        int indexOfTimeIntervals = getIndexOfTimeIntervals(standardTime);
        for(int i = 0; i < indexOfTimeIntervals; i ++) {
            timeIn2N0 += ratiosOfNiOverN0[i] * (timeIntervals[i+1] - timeIntervals[i]);
        }
        timeIn2N0 += ratiosOfNiOverN0[indexOfTimeIntervals] * (standardTime - timeIntervals[indexOfTimeIntervals]);

        return timeIn2N0;
    }

    @Override
    public void setParameters(double[] timeIntervals, double[] ratiosOfNiOverN0, int[] changeTypes) {
        // TODO Auto-generated method stub
        throw new IllegalArgumentException();
    }

    @Override
    public int getIndexOfTimeIntervalsInStepStone(double time) {

        return getIndexOfTimeIntervalsInStepStone(time, timeIntervals);

    }

    @Override
    public int getIndexOfTimeIntervalsInStepStone(double time, double[] timeIntervals) {

        if (timeIntervals[currentIntervalIndex] <= time) {
            if (currentIntervalIndex < numOfIntervals - 1) {

                if (time < timeIntervals[currentIntervalIndex + 1]) {
                    return currentIntervalIndex;
                }
                else {
                    currentIntervalIndex ++;
                    return getIndexOfTimeIntervalsInStepStone(time, timeIntervals);
                }
            }
            else {
                return currentIntervalIndex;
            }
        }
        else {
            if (timeIntervals[currentIntervalIndex - 1] <= time) {
                currentIntervalIndex --;
                return currentIntervalIndex;
            }
            else {
                currentIntervalIndex --;
                return getIndexOfTimeIntervalsInStepStone(time, timeIntervals);
            }
        }

    }

}
