/*
 * @(#)QSortAlgorithm.java	1.3   29 Feb 1996 James Gosling
 *
 * Copyright (c) 1994-1996 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL or COMMERCIAL purposes and
 * without fee is hereby granted.
 * Please refer to the file http://www.javasoft.com/copy_trademarks.html
 * for further important copyright and trademark information and to
 * http://www.javasoft.com/licensing.html for further important
 * licensing information for the Java (tm) Technology.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  SUN
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 */

import java.util.*;

public class QSort implements java.io.Serializable {
    /** This is a generic version of C.A.R Hoare's Quick Sort
     * algorithm.  This will handle arrays that are already
     * sorted, and arrays with duplicate keys.<BR>
     *
     * If you think of a one dimensional array as going from
     * the lowest index on the left to the highest index on the right
     * then the parameters to this function are lowest index or
     * left and highest index or right.  The first time you call
     * this function it will be with the parameters 0, a.length - 1.
     *
     * @param a       an integer array
     * @param lo0     left boundary of array partition
     * @param hi0     right boundary of array partition
     */
    static void QuickSort(int a[], int lo0, int hi0) {
        int lo = lo0;
        int hi = hi0;
        int mid;
        
        if ( hi0 > lo0) {
          /* Arbitrarily establishing partition element as the midpoint of
           * the array.
           *
           * Note: mid may not be equal to a[ ( lo0 + hi0 ) / 2 ] after a few step because
           * the value could be changed by swap function.
           */
            mid = a[ ( lo0 + hi0 ) / 2 ];
            
            // loop through the array until indices cross
            while( lo <= hi ) {
              /* find the first element that is greater than or equal to
               * the partition element starting from the left Index.
               */
                while( ( lo < hi0 ) && ( a[lo] < mid ) )
                    ++lo;
                
              /* find an element that is smaller than or equal to
               * the partition element starting from the right Index.
               */
                while( ( hi > lo0 ) && ( a[hi] > mid ) )
                    --hi;
                
                // if the indexes have not crossed, swap
                if( lo <= hi ) {
                    swap(a, lo, hi);
                    
                    ++lo;
                    --hi;
                }
            }
            
          /* If the right index has not reached the left side of array
           * must now sort the left partition.
           */
            if( lo0 < hi )
                QuickSort( a, lo0, hi );
            
          /* If the left index has not reached the right side of array
           * must now sort the right partition.
           */
            if( lo < hi0 )
                QuickSort( a, lo, hi0 );
            
        }
    }
    
    static void QuickSort(double a[], int lo0, int hi0) {
        int lo = lo0;
        int hi = hi0;
        double mid;
        
        if ( hi0 > lo0) {
          /* Arbitrarily establishing partition element as the midpoint of
           * the array.
           *
           * Note: mid may not be equal to a[ ( lo0 + hi0 ) / 2 ] after a few step because
           * the value could be changed by swap function.
           */
            mid = a[ ( lo0 + hi0 ) / 2 ];
            // loop through the array until indices cross
            while( lo <= hi ) {
              /* find the first element that is greater than or equal to
               * the partition element starting from the left Index.
               */
                while( ( lo < hi0 ) && ( a[lo] < mid ) )
                    ++lo;
              /* find an element that is smaller than or equal to
               * the partition element starting from the right Index.
               */
                while( ( hi > lo0 ) && ( a[hi] > mid ) )
                    --hi;
                // if the indexes have not crossed, swap
                if( lo <= hi ) {
                    swap(a, lo, hi);
                    
                    ++lo;
                    --hi;
                }
            }
            
          /* If the right index has not reached the left side of array
           * must now sort the left partition.
           */
            if( lo0 < hi )
                QuickSort( a, lo0, hi );
            
          /* If the left index has not reached the right side of array
           * must now sort the right partition.
           */
            if( lo < hi0 )
                QuickSort( a, lo, hi0 );
        }
    }
    
    static void QuickSort(List<Number> a, int lo0, int hi0) {
        int lo = lo0;
        int hi = hi0;
        double mid;
        
        if ( hi0 > lo0) {
          /* Arbitrarily establishing partition element as the midpoint of
           * the array.
           *
           * Note: mid may not be equal to a[ ( lo0 + hi0 ) / 2 ] after a few step because
           * the value could be changed by swap function.
           */
            mid = a.get(( lo0 + hi0 ) / 2).doubleValue();
            // loop through the array until indices cross
            while( lo <= hi ) {
              /* find the first element that is greater than or equal to
               * the partition element starting from the left Index.
               */
                while( ( lo < hi0 ) && ( a.get(lo).doubleValue() < mid ) )
                    ++lo;
              /* find an element that is smaller than or equal to
               * the partition element starting from the right Index.
               */
                while( ( hi > lo0 ) && ( a.get(hi).doubleValue() > mid ) )
                    --hi;
                // if the indexes have not crossed, swap
                if( lo <= hi ) {
                    swap(a, lo, hi);
                    
                    ++lo;
                    --hi;
                }
            }
            
          /* If the right index has not reached the left side of array
           * must now sort the left partition.
           */
            if( lo0 < hi )
                QuickSort( a, lo0, hi );
            
          /* If the left index has not reached the right side of array
           * must now sort the right partition.
           */
            if( lo < hi0 )
                QuickSort( a, lo, hi0 );
        }
    }
    
    private static final void swap(int a[], int i, int j) {
        int T;
        T = a[i];
        a[i] = a[j];
        a[j] = T;
    }
    
    private static final void swap(List<Number> a, int i, int j) {
    	Number T;
        T = a.get(i);
        a.set(i, a.get(j));
        a.set(j, T);
    }
    
    private static final void swap(double a[], int i, int j) {
        double T;
        T = a[i];
        a[i] = a[j];
        a[j] = T;
    }
    
    public static final void sort(int a[]) {
        QuickSort(a, 0, a.length - 1);
    }
    
    public static final void sort(int a[], int startIndex, int endIndex) {
        QuickSort(a, startIndex, endIndex);
    }
    
    public static final void sort(double a[]) {
        QuickSort(a, 0, a.length - 1);
    }
    
    public static final void sort(double a[], int startIndex, int endIndex) {
        QuickSort(a, startIndex, endIndex);
    }
    
    public static final void sort(List<Number> a) {
        QuickSort(a, 0, a.size() - 1);
    }
    
    public static final void sort(List<Number> a, int startIndex, int endIndex) {
        QuickSort(a, startIndex, endIndex);
    }
    
    public static void main(String[] args) {
    	/*
        int[] b = {157, 200, 154, 143, 160, 210, 151, 208, 148, 159, 206, 141, 158, 146, 161, 150, 204, 156, 145, 162, 140, 202, 144, 152, 155, 164, 142, 147, 153};
        
        System.out.println("before sorting... ");
        for (int k = 0; k < b.length; k ++) {
            System.out.print(b[k] + " ");
        }
        System.out.println();
        System.out.println("b.length = " + b.length);
        QSort.sort(b);
        
        System.out.println("after sorting... ");
        for (int k = 0; k < b.length; k ++) {
            System.out.print(b[k] + " ");
        }
        System.out.println();
        */
    	
        /*
        double[] a;
        int i;
        int size = 100;
        long startup_time;
        long sort_time;
        long end_time;
        java.text.DecimalFormat decimal = new java.text.DecimalFormat("#.###");
        
        if (args.length == 1) {
            size = Integer.parseInt(args[0]);
        }
        
        a  = new double[size];
        
        startup_time = java.lang.System.currentTimeMillis();
        
        for (i = 0; i < size; i++) {
            a[i] = Math.random() * 100;
        }
        
        sort_time = java.lang.System.currentTimeMillis();
        
        System.out.println("before sorting... ");
        for (int k = 0; k < a.length; k ++) {
            System.out.print(decimal.format(a[k]) + " ");
        }
        System.out.println();
        
        QSort.sort(a);
        
        System.out.println("after sorting... ");
        for (int k = 0; k < a.length; k ++) {
            System.out.print(decimal.format(a[k]) + " ");
        }
        System.out.println();
        
        for (int k = 1; k < a.length; k ++) {
            if (a[k - 1] > a[k]) {
                System.out.println("Wrong! k = " + k + " " + decimal.format(a[k - 1]) + " " + decimal.format(a[k]));
            }
        }
        
        end_time = java.lang.System.currentTimeMillis();
        java.lang.System.out.println("startup time: " + (sort_time - startup_time)/1000.0);
        java.lang.System.out.println("sort time: " + (end_time - sort_time)/1000.0);
         */
    	
    	int size = 10;
    	ArrayList<Number> values = new ArrayList(size);
    	for (int i = 0; i < size; i++) {
    		values.add(new Double(Math.random() * 100));
        }
    	QSort.sort(values);
    	
    	for (int i = 0; i < size; i++) {
    		System.out.println(values.get(i));
        }
    }
}


