import java.io.IOException;

import cern.jet.random.Poisson;
import cern.jet.random.engine.MersenneTwister64;

public class RextFitCoal_EstimatorForSinglePopModel extends EstimatorForSinglePopModel {
	protected int numOfExtraSearchedTimeIntervals = 2;
	
	public RextFitCoal_EstimatorForSinglePopModel() {
        super();
    }
	
	public RextFitCoal_EstimatorForSinglePopModel(int randSeed) {
        super(randSeed);
    }
	
	public PopSizeModelExponentialLinearInstant runOptimizeFreeTimes_NonAfricanPop(int[] obsSFS) throws IOException {
        setObsSFS(obsSFS);

        double obsTreeLength = 0;
        double[] obsBranchLengthsInGen = new double[obsSFS.length];
        int obsK = 0;
        for (int i = 0; i < obsSFS.length; i ++) {
            obsBranchLengthsInGen[i] = obsSFS[i] / mutRatePerGenome;
            obsK += obsSFS[i];
            obsTreeLength += obsBranchLengthsInGen[i];
        }

        table.setSampleSize(sampleSize);
        table.initialMatrix();

        RextFitCoal_MinFitnessOfNs optimizer = new RextFitCoal_MinFitnessOfNs();
        optimizer.setNumOfOmitStartSfs(numOfOmitStartSfs);
        optimizer.setNumOfOmitEndSfs(numOfOmitEndSfs);
        optimizer.setMutRatePerGenome(mutRatePerGenome);
        optimizer.setRandom(random);
        optimizer.setForbidInstantaneousGrowth(forbidInstantaneousGrowth);
        optimizer.setForbidInstantaneousDecline(forbidInstantaneousDecline);
        optimizer.setForbidExponentialGrowth(forbidExponentialGrowth);
        optimizer.setForbidExponentialDecline(forbidExponentialDecline);
        optimizer.setTimeUpperBound(timeUpperBound);
        optimizer.setGammaTimeLUS(gammaTimeLUS);
        optimizer.setGammaPopSizeLUS(gammaPopSizeLUS);
        optimizer.setMinChangeOfNPerPhase(minChangeOfNPerPhase);
        optimizer.setMaxChangeOfNPerPhase(maxChangeOfNPerPhase);
        optimizer.setMinRatioOfNOverN0(minRatioOfNOverN0);
        optimizer.setMaxRatioOfNOverN0(maxRatioOfNOverN0);
        
        if (useFoldedSFS) {
        	System.out.println("NonAfricanPop_EstimatorForSinglePopModel was written for analyzing special cases for Non-African populations, thus it does not support the unfolded SFS.");
        	System.out.println("Exit...");
        	System.exit(0);
        }

        optimizer.setObsSFS(obsSFS);

        int[] usedSFS = new int[obsSFS.length - numOfOmitStartSfs - numOfOmitEndSfs];
        for (int i = 0; i < usedSFS.length; i++) {
            usedSFS[i] = obsSFS[i + numOfOmitStartSfs];
        }

        Result4SinglePopHistory bestResult = null;

        int maxNumOfIntervalsSearched;
        if (fixedNumOfIntervals == 0){
            maxNumOfIntervalsSearched = sampleSize;
            for (int numOfIntervals = 1; numOfIntervals <= maxNumOfIntervalsSearched; numOfIntervals++) {
                Result4SinglePopHistory bestResultGivenNumOfParas = optimizer.optimize(numOfIntervals, table);
                for (int repeatIndex = 1; repeatIndex < numOfRepeats; repeatIndex ++) {
                    Result4SinglePopHistory oneRepeat = optimizer.optimize(numOfIntervals, table);
                    if (oneRepeat.fitness < bestResultGivenNumOfParas.fitness) {
                        bestResultGivenNumOfParas = oneRepeat;
                    }
                }
                System.out.println("numOfIntervals = " + numOfIntervals);
                printResult(bestResultGivenNumOfParas);

                if (bestResult == null) {
                    bestResult = bestResultGivenNumOfParas;
                } else {
                    if (bestResult.fitness < bestResultGivenNumOfParas.fitness * (likelihoodRate / 100 + 1)) {
                        break;
                    } else {
                        bestResult = bestResultGivenNumOfParas;
                    }
                }
            }
        } else {
            bestResult = optimizer.optimize(fixedNumOfIntervals, table);
            for (int repeatIndex = 1; repeatIndex < numOfRepeats; repeatIndex ++) {
                Result4SinglePopHistory oneRepeat = optimizer.optimize(fixedNumOfIntervals, table);
                if (oneRepeat.fitness < bestResult.fitness) {
                    bestResult = oneRepeat;
                }
            }
        }
        System.out.println("******** Best: ********");
        System.out.println("numOfIntervals = " + bestResult.popSizeChangeTypes.length);

        printResult(bestResult);
        
        System.out.println("numOfExtraSearchedTimeIntervals = " + numOfExtraSearchedTimeIntervals);
        
        Result4SinglePopHistory bestBottleneck = optimizer.optimize_NonAfricanPop(bestResult, numOfExtraSearchedTimeIntervals, table);
        for (int repeatIndex = 1; repeatIndex < numOfRepeats; repeatIndex ++) {
        	System.out.println("repeatIndex = " + repeatIndex);
        	
        	Result4SinglePopHistory bottleneck01 = optimizer.optimize_NonAfricanPop(bestResult, numOfExtraSearchedTimeIntervals, table);
        	if (bottleneck01.fitness < bestBottleneck.fitness) {
        		bestBottleneck = bottleneck01;
            }
        }
        System.out.println("******** Best bottleneck ********");
        System.out.println("numOfIntervals = " + bestBottleneck.popSizeChangeTypes.length);
        printResult(bestBottleneck);
        
        PopSizeModelExponentialLinearInstant popModel = new PopSizeModelExponentialLinearInstant();
        popModel.setParameters(bestBottleneck.startTimeOfIntervals, bestBottleneck.ratiosOfNiOverN0, bestBottleneck.popSizeChangeTypes);
        popModel.setN0(bestBottleneck.N0);

        return popModel;
    }
}
