
public class RextFitCoal_MinFitnessOfNs extends MinFitnessOfNs {
	
	// The inference upper bound for non-African populations is 2.0.
	// However, to infer the demographic changes beyond the last change, we have to extend this upper bound.
	protected double timeUpperBound4ExtraInference = 3.5;
	
	// Doesn't not support the folded SFS
	// Only consider the instant change beyond the last change.
	public Result4SinglePopHistory optimize_NonAfricanPop(Result4SinglePopHistory preHistory, int numOfSearchedTimeIntervals, TableForLinearModel table){
		
		// To identify the fixed history before the time of threshold
		int[] preHistoryChangeTypes   = preHistory.popSizeChangeTypes.clone();
		preHistoryChangeTypes[preHistoryChangeTypes.length - 1] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_INSTANT;
		
		double[] preHistoryStartTimes = preHistory.startTimeOfIntervals.clone();
		double[] preHistoryRatios     = preHistory.ratiosOfNiOverN0.clone();
		
		int thresholdIndex = preHistoryStartTimes.length - 1;
		double thresholdTime = preHistoryStartTimes[preHistoryStartTimes.length - 1];
		
		final int totalNumOfTimeIntervals = thresholdIndex + 1 + numOfSearchedTimeIntervals;
		
		int[] currentChangeTypes = new int[totalNumOfTimeIntervals];
        double[] currentStartTimes = new double[totalNumOfTimeIntervals];
        double[] currentRatios     = new double[totalNumOfTimeIntervals];
        
        double[] bestStartTimes    = new double[totalNumOfTimeIntervals];
        double[] bestRatios        = new double[totalNumOfTimeIntervals];

        int maxIterations = NUM_BASIC_ITERATIONS * (totalNumOfTimeIntervals - 1) + 1;
		
        // To fix the estimated history before the time of threshold
        for (int i = 0; i <= thresholdIndex; i ++) {
        	currentChangeTypes[i] = preHistoryChangeTypes[i];
        	currentStartTimes[i]  = preHistoryStartTimes[i];
        	currentRatios[i]      = preHistoryRatios[i]; // The last ratio is a variable to be estimated.
        	
        	bestStartTimes[i]     = preHistoryStartTimes[i];
        	bestRatios[i]         = preHistoryRatios[i]; // The last ratio is a variable to be estimated.
        }
        
        double oneAvgTimeLength = (timeUpperBound4ExtraInference - thresholdTime) / (numOfSearchedTimeIntervals + 1);
        for (int i = thresholdIndex + 1; i < totalNumOfTimeIntervals; i ++) {
        	currentChangeTypes[i] = PopSizeModelExponentialLinearInstant.CHANGE_TYPE_INSTANT;
        	currentRatios[i] = 1.0d;
        	currentStartTimes[i] = thresholdTime + oneAvgTimeLength * (i - thresholdIndex);
        	
        	bestRatios[i] = 1.0d;
        	bestStartTimes[i] = currentStartTimes[i];
        }
        
        double log10MinRatioOfNOverN0 = Math.log10(minRatioOfNOverN0);
        double log10MaxRatioOfNOverN0 = Math.log10(maxRatioOfNOverN0);
        double[] logSFS = initialLogSFSFactorial(obsSFS);
        
        double rPopSize = 1.0;
        double qPopSize = Math.pow(2.0, -1.0 / (1 + (numOfSearchedTimeIntervals - 1) * gammaPopSizeLUS));
        double rStartTimes = timeUpperBound4ExtraInference;
        double qStartTimes = Math.pow(2.0, -1.0 / (1 + (numOfSearchedTimeIntervals - 1) * gammaPopSizeLUS));

        Result4SinglePopHistory bestResult = new Result4SinglePopHistory();
        bestResult.N0 = preHistory.N0;
        bestResult.fitness = preHistory.fitness;
        bestResult.ratiosOfNiOverN0 = currentRatios.clone();
        bestResult.startTimeOfIntervals = currentStartTimes.clone();
        bestResult.popSizeChangeTypes = currentChangeTypes.clone();
        
        double minFitness = Double.MAX_VALUE;

        for (int iteration = 1; iteration <= maxIterations; iteration++) {
            int circleCount = 0;
            boolean feasible = true;
            currentRatios[0] = 1.0d;
            do {
            	// The last ratio in the previously estimated demography is a variable to be estimated. But the time remains un-changed.
                for (int i = thresholdIndex; i < totalNumOfTimeIntervals; i ++) {
                    double rand = sampleBounded(Math.log10(bestRatios[i]), rPopSize, log10MinRatioOfNOverN0, log10MaxRatioOfNOverN0);
                    currentRatios[i] = Math.pow(10, rand);
                }
                feasible = examineFeasiblePopulation(totalNumOfTimeIntervals, currentRatios, currentChangeTypes);
                circleCount++;
            } while (circleCount < MAX_RAND_SAMPLING_CIRCLE_COUNT && (!feasible));

            if (!feasible) {
                continue;
            }

            circleCount = 0;
            currentStartTimes[0] = 0.0d;
            do {
            	// getRandomSelectedStartTimes_NonAfricanPop(double thresholdTime, int thresholdIndex, int totalNumOfTimeIntervals, double[] initialTimePoints, double range
                currentStartTimes = getRandomSelectedStartTimes_NonAfricanPop(thresholdTime, thresholdIndex, totalNumOfTimeIntervals, bestStartTimes, rStartTimes);
                feasible = examineFeasibleOfStartTimes(currentStartTimes, numOfSearchedTimeIntervals);
                circleCount++;
            } while (circleCount < MAX_RAND_SAMPLING_CIRCLE_COUNT && (!feasible));

            if (!feasible) {
                continue;
            }
            PopSizeModelExponentialLinearInstant popModel = new PopSizeModelExponentialLinearInstant();
            popModel.setParameters(currentStartTimes, currentRatios, currentChangeTypes.clone());

            double[] tabulatedTimes = table.startTimeOfDetailedIntervals;
            PopSizeModelInstantChange convertedModel = popModel.convert(tabulatedTimes);
            double[] expectedBranchLengthsIn2N0 = table.getExpectedBranchLengths(convertedModel);

            // numOfOmitStartSfs: 0-based, inclusive
            // endIndex: 0-based, inclusive
            double obsMutSum = sum(obsSFS, numOfOmitStartSfs, obsSFS.length - numOfOmitEndSfs - 1);
            double expectedBranchLengthsIn2N0Sum = sum(expectedBranchLengthsIn2N0, numOfOmitStartSfs, obsSFS.length - numOfOmitEndSfs - 1);

            double thetaScaledIn2N0 = obsMutSum / expectedBranchLengthsIn2N0Sum;
            double N0 = thetaScaledIn2N0 / mutRatePerGenome * 0.5;
            feasible = examineIntervalOfStartTimes(currentStartTimes, N0);
            if (!feasible) {
                continue;
            }
            
            // initialSize: 1-based, inclusive
            // endSize: 1-based, inclusive
            double localFitness = estimateFitness(thetaScaledIn2N0, expectedBranchLengthsIn2N0, obsSFS, logSFS, 1 + numOfOmitStartSfs, obsSFS.length - numOfOmitEndSfs);

            if (minFitness > localFitness) {
                minFitness = localFitness;
                bestResult.N0 = N0;
                bestResult.fitness = localFitness;
                bestResult.popSizeChangeTypes = currentChangeTypes.clone();
                bestResult.startTimeOfIntervals = currentStartTimes.clone();
                bestResult.ratiosOfNiOverN0 = currentRatios.clone();

                double[] temp = bestStartTimes;
                bestStartTimes = currentStartTimes;
                currentStartTimes = temp;

                temp = bestRatios;
                bestRatios = currentRatios;
                currentRatios = temp;

            } else {
                rStartTimes *= qStartTimes;
                rPopSize *= qPopSize;
                if ((rStartTimes < PRECISION_THRESHOLD) && (rPopSize < PRECISION_THRESHOLD)) {
                    break;
                }
            }
        }
        
        return bestResult;
    }
	
	protected double[] getRandomSelectedStartTimes_NonAfricanPop(double thresholdTime, int thresholdIndex, int totalNumOfTimeIntervals, double[] initialTimePoints, double range) {

        double[] startTimes = new double[totalNumOfTimeIntervals];
        for (int i = 0; i <= thresholdIndex; i ++) {
        	startTimes[i] = initialTimePoints[i]; // Within the fixed demographic history
        }

        double minTime = thresholdTime + MIN_SPANNING_TIME_4_ONE_PHASE;
        double maxTime = timeUpperBound4ExtraInference - MIN_SPANNING_TIME_4_ONE_PHASE;

        for (int i = thresholdIndex + 1; i < totalNumOfTimeIntervals; i ++) {
            double initialPoint = initialTimePoints[i];
            startTimes[i] = sampleBounded(initialPoint, range, minTime, maxTime);
        }
        QSort.sort(startTimes);

        return startTimes;
    }
}
