import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.List;

import org.kohsuke.args4j.CmdLineParser;

public class RextFitCoal_SinglePopDecoder extends SinglePopDecoder {

	public void run_NonAfricanPop(SinglePopDecoderInput singlePopDecoderInput) throws Exception {
        String tablesDir = singlePopDecoderInput.getTablesDir();
        String sfsFileName = singlePopDecoderInput.getSfsFileName();
        String outputFileName = singlePopDecoderInput.getOutputFileName();
        double generationTime = singlePopDecoderInput.getGenerationTime();
        double mutRatePerKb = singlePopDecoderInput.getMutRatePerKb();
        double lengthOfGenome = singlePopDecoderInput.getLengthOfGenome();
        double yearMax = singlePopDecoderInput.getYearMax();
        double yearMin = singlePopDecoderInput.getYearMin();
        boolean forbidInstantaneousGrowth = singlePopDecoderInput.isForbidInstantaneousGrowth();
        boolean forbidExponentialGrowth = singlePopDecoderInput.isForbidExponentialGrowth();
        boolean forbidInstantaneousDecline = singlePopDecoderInput.isForbidInstantaneousDecline();
        boolean forbidExponentialDecline = singlePopDecoderInput.isForbidExponentialDecline();
        int fixedNumOfIntervals = singlePopDecoderInput.getFixedNumOfIntervals();
        double likelihoodRate = singlePopDecoderInput.getLikelihoodRate();
        int numOfRepeats = singlePopDecoderInput.getNumOfRepeats();
        int numOfOmitStartSfs = singlePopDecoderInput.getNumOfOmitStartSfs();
        int numOfOmitEndSfs = singlePopDecoderInput.getNumOfOmitEndSfs();
        double timeUpperBound = singlePopDecoderInput.getTimeUpperBound();
        double gammaTimeLUS = singlePopDecoderInput.getGammaTimeLUS();
        double gammaPopSizeLUS = singlePopDecoderInput.getGammaPopSizeLUS();
        int randSeed = singlePopDecoderInput.getRandSeed();
        double minChangeOfNPerPhase = singlePopDecoderInput.getMinChangeOfNPerPhase();
        double maxChangeOfNPerPhase = singlePopDecoderInput.getMaxChangeOfNPerPhase();
        double minRatioOfNOverN0 = singlePopDecoderInput.getMinRatioOfNOverN0();
        double maxRatioOfNOverN0 = singlePopDecoderInput.getMaxRatioOfNOverN0();
        double mutRatePerGenome = mutRatePerKb * lengthOfGenome;

        TableForLinearModel.tableDir = tablesDir;

        if (forbidExponentialDecline && forbidExponentialGrowth && forbidInstantaneousDecline && forbidInstantaneousGrowth) {
            fixedNumOfIntervals = 1;
        }

        RextFitCoal_EstimatorForSinglePopModel estimator;
        if (randSeed == 0) {
            estimator = new RextFitCoal_EstimatorForSinglePopModel();
        } else {
            estimator = new RextFitCoal_EstimatorForSinglePopModel(randSeed);
        }

        estimator.setMutRatePerGenome(mutRatePerGenome);
        estimator.setForbidInstantaneousGrowth(forbidInstantaneousGrowth);
        estimator.setForbidExponentialGrowth(forbidExponentialGrowth);
        estimator.setForbidInstantaneousDecline(forbidInstantaneousDecline);
        estimator.setForbidExponentialDecline(forbidExponentialDecline);
        estimator.setFixedNumOfIntervals(fixedNumOfIntervals);
        estimator.setLikelihoodRate(likelihoodRate);
        estimator.setNumOfRepeats(numOfRepeats);
        estimator.setNumOfOmitStartSfs(numOfOmitStartSfs);
        estimator.setNumOfOmitEndSfs(numOfOmitEndSfs);
        estimator.setTimeUpperBound(timeUpperBound);
        estimator.setGammaTimeLUS(gammaTimeLUS);
        estimator.setGammaPopSizeLUS(gammaPopSizeLUS);
        estimator.setMinChangeOfNPerPhase(minChangeOfNPerPhase);
        estimator.setMaxChangeOfNPerPhase(maxChangeOfNPerPhase);
        estimator.setMinRatioOfNOverN0(minRatioOfNOverN0);
        estimator.setMaxRatioOfNOverN0(maxRatioOfNOverN0);

        FileReader fileReader = new FileReader(sfsFileName);
        LineNumberReader lineNumberReader = new LineNumberReader(fileReader);

        String line;
        int dataIndex = 0;
        File outputFile = new File(outputFileName + ".txt");
        FileOutputStream out = new FileOutputStream(outputFile);
        out.write("year\tpopSize\n".getBytes());

        List<int[]> obsSFSs = new ArrayList<>();
        List<Integer> sampleSizes = new ArrayList<>();
        boolean missingDataType = false;

        while (true) {
            line = lineNumberReader.readLine();
            String[] listOfWords = null;
            boolean fileEnd = false;
            if (line == null) {
                fileEnd = true;
                listOfWords = new String[1];
                listOfWords[0] = "no data";
            } else {
                listOfWords = line.split("\\s+");
            }
            if ((line == null && missingDataType) || "SAMPLESIZE".equals(listOfWords[0].toUpperCase())) {
                missingDataType = true;

                System.out.println("NonAfricanPop_SinglePopDecoder was written for analyzing special cases for Non-African populations, thus it does not support this type of data.");

            } else {
                if (fileEnd) {
                    break;
                }
                dataIndex++;
                System.out.println("Data:" + dataIndex);

                if (listOfWords.length != 0) {
                    int[] obsSFS = new int[listOfWords.length];
                    for (int i = 0; i < listOfWords.length; i++) {
                        obsSFS[i] = Integer.parseInt(listOfWords[i]);
                    }

                    PopSizeModelExponentialLinearInstant demography = estimator.runOptimizeFreeTimes_NonAfricanPop(obsSFS);
                    String result = outputPopStr(demography, generationTime, yearMax, yearMin) + "\n";
                    out.write(result.getBytes());

                }
            }
            if (line == null) {
                break;
            }
        }

        out.close();
        lineNumberReader.close();
    }
	
	public static void main(String[] args) throws Exception {

        SinglePopDecoderInput singlePopDecoderInput = new SinglePopDecoderInput();
        CmdLineParser parser = new CmdLineParser(singlePopDecoderInput);
        parser.parseArgument(args);

        RextFitCoal_SinglePopDecoder singlePopDecoder = new RextFitCoal_SinglePopDecoder();
        long startTime01 = System.currentTimeMillis();
        singlePopDecoder.run_NonAfricanPop(singlePopDecoderInput);
        
        long endTime01 = System.currentTimeMillis();
        double time = (endTime01 - startTime01) / 1000.0;
        System.out.println("computing time = " + String.format("%.2f", time) + " seconds");
    }
}
