import org.kohsuke.args4j.*;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

public class SinglePopDecoder {

    public void run(SinglePopDecoderInput singlePopDecoderInput) throws Exception {
        String tablesDir = singlePopDecoderInput.getTablesDir();
        String sfsFileName = singlePopDecoderInput.getSfsFileName();
        String outputFileName = singlePopDecoderInput.getOutputFileName();
        
        double generationTime = singlePopDecoderInput.getGenerationTime();
        double mutRatePerKb = singlePopDecoderInput.getMutRatePerKb();
        double lengthOfGenome = singlePopDecoderInput.getLengthOfGenome();
        double yearMax = singlePopDecoderInput.getYearMax();
        double yearMin = singlePopDecoderInput.getYearMin();
        boolean forbidInstantaneousGrowth = singlePopDecoderInput.isForbidInstantaneousGrowth();
        boolean forbidExponentialGrowth = singlePopDecoderInput.isForbidExponentialGrowth();
        boolean forbidInstantaneousDecline = singlePopDecoderInput.isForbidInstantaneousDecline();
        boolean forbidExponentialDecline = singlePopDecoderInput.isForbidExponentialDecline();
        int fixedNumOfIntervals = singlePopDecoderInput.getFixedNumOfIntervals();
        double likelihoodRate = singlePopDecoderInput.getLikelihoodRate();
        
        int numOfRepeats = singlePopDecoderInput.getNumOfRepeats();
        int numOfOmitStartSfs = singlePopDecoderInput.getNumOfOmitStartSfs();
        int numOfOmitEndSfs = singlePopDecoderInput.getNumOfOmitEndSfs();
        int numOfCollapsedEndSfs = singlePopDecoderInput.getNumOfCollapsedEndSfs();
        
        double timeUpperBound = singlePopDecoderInput.getTimeUpperBound();
        double gammaTimeLUS = singlePopDecoderInput.getGammaTimeLUS();
        double gammaPopSizeLUS = singlePopDecoderInput.getGammaPopSizeLUS();
        int randSeed = singlePopDecoderInput.getRandSeed();
        double minChangeOfNPerPhase = singlePopDecoderInput.getMinChangeOfNPerPhase();
        double maxChangeOfNPerPhase = singlePopDecoderInput.getMaxChangeOfNPerPhase();
        double minRatioOfNOverN0 = singlePopDecoderInput.getMinRatioOfNOverN0();
        double maxRatioOfNOverN0 = singlePopDecoderInput.getMaxRatioOfNOverN0();
        double mutRatePerGenome = mutRatePerKb * lengthOfGenome;
        
        boolean useFoldedSFS = singlePopDecoderInput.getUseFoldedSFS(); // Haipeng
        
        if (useFoldedSFS) {
        	System.out.println("Note: folded SFS will be used.");
        }

        TableForLinearModel.tableDir = tablesDir;

        if (forbidExponentialDecline && forbidExponentialGrowth && forbidInstantaneousDecline && forbidInstantaneousGrowth) {
            fixedNumOfIntervals = 1;
        }

        EstimatorForSinglePopModel estimator;
        if (randSeed == 0) {
            estimator = new EstimatorForSinglePopModel();
        } else {
            estimator = new EstimatorForSinglePopModel(randSeed);
        }

        estimator.setMutRatePerGenome(mutRatePerGenome);
        estimator.setForbidInstantaneousGrowth(forbidInstantaneousGrowth);
        estimator.setForbidExponentialGrowth(forbidExponentialGrowth);
        estimator.setForbidInstantaneousDecline(forbidInstantaneousDecline);
        estimator.setForbidExponentialDecline(forbidExponentialDecline);
        estimator.setFixedNumOfIntervals(fixedNumOfIntervals);
        estimator.setLikelihoodRate(likelihoodRate);
        
        estimator.setNumOfRepeats(numOfRepeats);
        estimator.setNumOfOmitStartSfs(numOfOmitStartSfs);
        estimator.setNumOfOmitEndSfs(numOfOmitEndSfs);
        estimator.setNumOfCollapsedEndSfs(numOfCollapsedEndSfs);
        
        estimator.setTimeUpperBound(timeUpperBound);
        estimator.setGammaTimeLUS(gammaTimeLUS);
        estimator.setGammaPopSizeLUS(gammaPopSizeLUS);
        estimator.setMinChangeOfNPerPhase(minChangeOfNPerPhase);
        estimator.setMaxChangeOfNPerPhase(maxChangeOfNPerPhase);
        estimator.setMinRatioOfNOverN0(minRatioOfNOverN0);
        estimator.setMaxRatioOfNOverN0(maxRatioOfNOverN0);
        
        estimator.setUseFoldedSFS(useFoldedSFS); // haipeng

        FileReader fileReader = new FileReader(sfsFileName);
        LineNumberReader lineNumberReader = new LineNumberReader(fileReader);

        String line;
        int dataIndex = 0;
        File outputFile = new File(outputFileName + ".txt");
        FileOutputStream out = new FileOutputStream(outputFile);
        out.write("year\tpopSize\n".getBytes());

        List<int[]> obsSFSs = new ArrayList<>();
        List<Integer> sampleSizes = new ArrayList<>();
        boolean missingDataType = false;

        while (true) {
            line = lineNumberReader.readLine();
            if (line == null) {
                break;
            }
            
            line = line.trim();
            if (line.length() > 0) { // Added by Haipeng, the if ().
            	String[] listOfWords = null;
                boolean fileEnd = false;
                if (line == null) {
                    fileEnd = true;
                    listOfWords = new String[1];
                    listOfWords[0] = "no data";
                } else {
                    listOfWords = line.split("\\s+");
                }
                if ((line == null && missingDataType) || "SAMPLESIZE".equals(listOfWords[0].toUpperCase())) {
                    missingDataType = true;

                    if (line == null || line.length() == 0) {
                        dataIndex++;
                        System.out.println("Data: " + dataIndex);
                        PopSizeModelExponentialLinearInstant demography = estimator.runOptimizeFreeTimes(obsSFSs, sampleSizes);
                        String result = outputPopStr(demography, generationTime, yearMax, yearMin) + "\n";
                        out.write(result.getBytes());

                    } else {
                        int sampleSize = Integer.parseInt(listOfWords[1]);
                        if (sampleSize + 1 != listOfWords.length) {
                            throw new Exception("SampleSize error!");
                        }
                        sampleSizes.add(sampleSize);
                        int[] obsSFS = new int[listOfWords.length - 2];
                        for (int i = 2; i < listOfWords.length; i++) {
                            obsSFS[i - 2] = Integer.parseInt(listOfWords[i]);
                        }
                        obsSFSs.add(obsSFS);
                    }

                } else {
                    if (fileEnd) {
                        break;
                    }
                    dataIndex++;
                    System.out.println("Data:" + dataIndex);

                    if (listOfWords.length != 0) {
                        int[] obsSFS = new int[listOfWords.length];
                        for (int i = 0; i < listOfWords.length; i++) {
                            obsSFS[i] = Integer.parseInt(listOfWords[i]);
                        }

                        PopSizeModelExponentialLinearInstant demography = estimator.runOptimizeFreeTimes(obsSFS);
                        String result = outputPopStr(demography, generationTime, yearMax, yearMin) + "\n";
                        out.write(result.getBytes());

                    }
                }
                
            }
        }

        out.close();
        lineNumberReader.close();
    }

    protected  String outputPopStr(PopSizeModelExponentialLinearInstant demography, double generationTime, double maxTimeYear, double minTimeYear) {
        double N0 = demography.getN0();
        double[] timeIntervals = demography.getTimeIntervals();
        StringBuilder log = new StringBuilder();

        if (demography.getNumOfIntervals() == 1) {
            double popSize = demography.getRatioOfNiOverN0(timeIntervals[0]) * N0;
            log.append(Math.round(minTimeYear)).append("\t");
            log.append(Math.round(popSize)).append("\n");
            log.append(Math.round(maxTimeYear)).append("\t");
            log.append(Math.round(popSize)).append("\n");
        } else {
            int startIndex = 0;
            int endIndex = 0;
            boolean checkStart = false;
            for (int i = 0; i < demography.getNumOfIntervals(); i++) {
                double timeYear = demography.calculateTimeIn2N0(timeIntervals[i]) * 2 * N0 * generationTime;
                if (timeYear >= minTimeYear) {
                    startIndex = i - 1;
                    checkStart = true;
                    break;
                }
            }

            for (int i = 0; i < demography.getNumOfIntervals(); i++) {
                double timeYear = demography.calculateTimeIn2N0(timeIntervals[i]) * 2 * N0 * generationTime;
                if (timeYear > maxTimeYear) {
                    endIndex = i;
                    break;
                }
            }
            if (endIndex == 0) {
                endIndex = demography.getNumOfIntervals();
            }

            if (checkStart) {
                for (int i = startIndex; i < endIndex; i++) {
                    int changeType = demography.getChangeTypes()[i];
                    double timeYear = demography.calculateTimeIn2N0(timeIntervals[i]) * 2 * N0 * generationTime;
                    double popSize = demography.getRatioOfNiOverN0(timeIntervals[i]) * N0;
                    switch (changeType) {
                        case PopSizeModelExponentialLinearInstant.CHANGE_TYPE_INSTANT:
                            if (timeYear < minTimeYear) {
                                timeYear = minTimeYear;
                            }
                            log.append(Math.round(timeYear)).append("\t");
                            log.append(Math.round(popSize)).append("\n");
                            if (i == endIndex - 1) {
                                log.append(Math.round(maxTimeYear)).append("\t");
                                log.append(Math.round(popSize)).append("\n");
                            } else {
                                log.append(Math.round(demography.calculateTimeIn2N0(timeIntervals[i + 1]) * 2 * N0 * generationTime)).append("\t");
                                log.append(Math.round(popSize)).append("\n");
                            }
                            break;
                        case PopSizeModelExponentialLinearInstant.CHANGE_TYPE_LAST_CONSTANT:
                            log.append(Math.round(timeYear)).append("\t");
                            log.append(Math.round(popSize)).append("\n");
                            log.append(Math.round(maxTimeYear)).append("\t");
                            log.append(Math.round(popSize)).append("\n");
                            break;
                        case PopSizeModelExponentialLinearInstant.CHANGE_TYPE_EXPONENTIAL:
                            double numOfPieces = 100;
                            double nextIntervalTimeYear = demography.calculateTimeIn2N0(timeIntervals[i + 1]) * 2 * N0 * generationTime;
                            double nextIntervalPopSize = demography.getRatioOfNiOverN0(timeIntervals[i + 1]) * N0;
                            if (nextIntervalTimeYear > maxTimeYear) {
                                double rate = Math.exp(Math.log(nextIntervalPopSize/popSize)/(nextIntervalTimeYear - timeYear) * (maxTimeYear - timeYear));
                                nextIntervalPopSize = rate * popSize;
                                nextIntervalTimeYear = maxTimeYear;
                            }
                            if (timeYear < minTimeYear) {
                                double rate = Math.exp(Math.log(nextIntervalPopSize/popSize)/(nextIntervalTimeYear - timeYear) * (minTimeYear - timeYear));
                                popSize =  rate * popSize;
                                timeYear = minTimeYear;
                            }
                            double eachStepTimeLength = (nextIntervalTimeYear - timeYear)/numOfPieces;
                            double eachStepPopRate = Math.exp(Math.log(nextIntervalPopSize/popSize)/numOfPieces);
                            double currentTimeYear = timeYear;
                            double currentPopSize = popSize;
                            for(int j = 0; j < numOfPieces; j++ ) {
                                log.append(Math.round(currentTimeYear)).append("\t");
                                currentTimeYear += eachStepTimeLength;
                                log.append(Math.round(currentPopSize)).append("\n");
                                currentPopSize *= eachStepPopRate;
                            }
                            if (i == endIndex - 1) {
                                log.append(Math.round(maxTimeYear)).append("\t");
                                log.append(Math.round(nextIntervalPopSize)).append("\n");
                            }
                            break;
                        default:
                    }
                }
            } else {
                double popSize = demography.getRatioOfNiOverN0(timeIntervals[endIndex - 1]) * N0;
                log.append(Math.round(minTimeYear)).append("\t");
                log.append(Math.round(popSize)).append("\n");
                log.append(Math.round(maxTimeYear)).append("\t");
                log.append(Math.round(popSize)).append("\n");
            }
        }



        return log.toString();
    }

    public static void main(String[] args) throws Exception {

        SinglePopDecoderInput singlePopDecoderInput = new SinglePopDecoderInput();
        CmdLineParser parser = new CmdLineParser(singlePopDecoderInput);
        parser.parseArgument(args);

        SinglePopDecoder SinglePopDecoder = new SinglePopDecoder();
        long startTime01 = System.currentTimeMillis();
        SinglePopDecoder.run(singlePopDecoderInput);
        long endTime01 = System.currentTimeMillis();
        double time = (endTime01 - startTime01) / 1000.0;
        System.out.println("computing time = " + String.format("%.2f", time) + " seconds");
    }
}
