import org.kohsuke.args4j.*;

public class SinglePopDecoderInput {
    @Option(required = true, name = "-table", usage = "Directory containing table files")
    private String tablesDir;

    @Option(required = true, name = "-input", usage = "Your SFS file")
    private String sfsFileName;

    @Option(required = true, name = "-output", usage = "The prefix of your output files")
    private String outputFileName;

    @Option(required = true, name = "-mutationRate", usage = "Mutation rate PER KB")
    private double mutRatePerKb;

    @Option(required = true, name = "-generationTime", usage = "Generation time")
    private double generationTime;

    @Option(required = true, name = "-genomeLength", usage = "The length of your genome IN KB")
    private double lengthOfGenome;

    @Option(name = "-yearMax", usage = "The max year you want to plot")
    private int yearMax = 4000000;

    @Option(name = "-yearMin", usage = "The min year you want to plot")
    private int yearMin = 2000;

    @Option(name = "-noIG", usage = "Forbid instantaneous population growth")
    private boolean forbidInstantaneousGrowth = false;

    @Option(name = "-noEG", usage = "Forbid exponential population growth")
    private boolean forbidExponentialGrowth = false;

    @Option(name = "-noID", usage = "Forbid instantaneous population decline")
    private boolean forbidInstantaneousDecline = false;

    @Option(name = "-noED", usage = "Forbid exponential population decline")
    private boolean forbidExponentialDecline = false;

    @Option(name = "-numOfIntervals", usage = "Fixed number of time intervals")
    private int fixedNumOfIntervals = 0;

    @Option(name = "-logLPRate", usage = "Log-likelihood promotion rate")
    private double likelihoodRate = 20;

    @Option(name = "-repeats", usage = "Number of estimate repeats")
    private int numOfRepeats = 100;

    @Option(name = "-omitStartSFS", usage = "Number of omitted start types of SFS")
    private int numOfOmitStartSfs = 0;

    @Option(name = "-omitEndSFS", usage = "Number of omitted end types of SFS")
    private int numOfOmitEndSfs = 0;
    
    @Option(name = "-collapsedEndSFS", usage = "Collapsing the last (collapsedEndSFS) SFS categories into one class")
    private int numOfCollapsedEndSfs = 0;

	@Option(name = "-randSeed", usage = "Seed of Random number")
    private int randSeed = 0;

    @Option(name = "-timeUpperBound", usage = "Upper bound of estimated standard coalescent time")
    private double timeUpperBound = 2.0;

    @Option(name = "-gammaTimeLUS", usage = "Gamma parameter of LUS when sampling time")
    private double gammaTimeLUS = 600.0;

    @Option(name = "-gammaPopSizeLUS", usage = "Gamma parameter of LUS when sampling population size")
    private double gammaPopSizeLUS = 600.0;

    @Option(name = "-minChangePerPhase", usage = "Min change of N per time interval")
    private double minChangeOfNPerPhase = 0.01;

    @Option(name = "-maxChangePerPhase", usage = "Max change of N per time interval")
    private double maxChangeOfNPerPhase = 100;

    @Option(name = "-minRatioOverN0", usage = "Min ratio of N over N0")
    private double minRatioOfNOverN0 = 0.001;

    @Option(name = "-maxRatioOverN0", usage = "Min ratio of N over N0")
    private double maxRatioOfNOverN0 = 1000;
    
    @Option(name = "-foldedSFS", usage = "true or false")
    private boolean useFoldedSFS = false; // Haipeng

    public SinglePopDecoderInput() {
    }

    public String getTablesDir() {
        return tablesDir;
    }

    public String getSfsFileName() {
        return sfsFileName;
    }

    public String getOutputFileName() {
        return outputFileName;
    }

    public double getMutRatePerKb() {
        return mutRatePerKb;
    }

    public double getGenerationTime() {
        return generationTime;
    }

    public double getLengthOfGenome() {
        return lengthOfGenome;
    }

    public int getYearMax() {
        return yearMax;
    }

    public int getYearMin() {
        return yearMin;
    }

    public boolean isForbidInstantaneousGrowth() {
        return forbidInstantaneousGrowth;
    }

    public boolean isForbidExponentialGrowth() {
        return forbidExponentialGrowth;
    }

    public boolean isForbidInstantaneousDecline() {
        return forbidInstantaneousDecline;
    }

    public boolean isForbidExponentialDecline() {
        return forbidExponentialDecline;
    }

    public int getFixedNumOfIntervals() {
        return fixedNumOfIntervals;
    }

    public double getLikelihoodRate() {
        return likelihoodRate;
    }

    public int getNumOfRepeats() {
        return numOfRepeats;
    }

    public int getNumOfOmitStartSfs() {
        return numOfOmitStartSfs;
    }

    public int getNumOfOmitEndSfs() {
        return numOfOmitEndSfs; 
    }
    
    // If SFS is truncated, this collapse option should not be used. Or the option will be ignored.
    public int getNumOfCollapsedEndSfs() {
		return numOfCollapsedEndSfs; // Haipeng
	}

    public int getRandSeed() {
        return randSeed;
    }

    public double getTimeUpperBound() {
        return timeUpperBound;
    }

    public double getGammaTimeLUS() {
        return gammaTimeLUS;
    }

    public double getGammaPopSizeLUS() {
        return gammaPopSizeLUS;
    }

    public double getMinChangeOfNPerPhase() {
        return minChangeOfNPerPhase;
    }

    public double getMaxChangeOfNPerPhase() {
        return maxChangeOfNPerPhase;
    }

    public double getMinRatioOfNOverN0() {
        return minRatioOfNOverN0;
    }

    public double getMaxRatioOfNOverN0() {
        return maxRatioOfNOverN0;
    }
    
    // The SFS should be used as the input. If useFoldedSFS is true, the SFS will be folded
    public boolean getUseFoldedSFS() {
        return useFoldedSFS; // Haipeng
    }
}
