import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

public class TableForLinearModel {
    public static String tableDir;
    public int sampleSize;
    protected double tenMinusTwenty = Math.pow(10, -20);
    protected double minProbability = 0.00000001;

    public double threshold_DeltaT = 0.001;
    public double maxTimeTracedBack = Double.MAX_VALUE;

    protected int numOfDetailedIntervals = 0;

    protected double reducedPercNumOfLineagesPerInterval = 1.0d - 1E-2;

    public double expectedTreeLength = 0.0d;

    protected double[] k_kMinusOne_dividedBy2 = null;

    protected double[] valuesArray;
    protected double[] valuesArray_empty;
    protected int sampleSizePlusOne;

    public boolean loaded = false;

    public double[] startTimeOfDetailedIntervals;
    protected double[] oneOverK = null;

    protected double[] expectedNumOfLineages;

    protected double[] branchLengths;
    protected double[][] detailedMatrix;
    protected double[] detailedMarginalValues;

    protected String nl = System.getProperty("line.separator");

    protected double[] logFactorials;

    public TableForLinearModel() {
        setSampleSize(sampleSize);
    }

    protected double getExpectedTreeLength(int sampleSize) {
        double treeLength = 0;
        for (double k = 1; k < sampleSize; k ++) {
            treeLength += 2.0d / k;
        }

        return treeLength;
    }

    public void initialMatrix() {
        int nMinusOne = sampleSize - 1;

        if (isTableLoadable()) {
            loadMatrix();
        }
        else {
            System.out.println("Table is initializing.");
            long startTime = System.currentTimeMillis();
            initialData();
            long endTime = System.currentTimeMillis();
            long cost = endTime - startTime;
            System.out.println("Initialized! Cost " + cost + "ms!");
            saveMatrix();
        }

        detailedMarginalValues = new double[numOfDetailedIntervals];
        for (int intervalIndex = 0; intervalIndex < numOfDetailedIntervals; intervalIndex ++) {
            double[] oneRow = detailedMatrix[intervalIndex];
            for (int i = 0; i < nMinusOne; i ++) {
                detailedMarginalValues[intervalIndex] += oneRow[i];
            }
        }
    }

    protected double[] calculateIntervalPoints(int sampleSize) {

        int estSize = (int)(Math.log(sampleSize) / (-Math.log(reducedPercNumOfLineagesPerInterval)) + 3);
        List<Double> listOfPoints = new ArrayList(estSize);

        double numOfLineages = sampleSize;
        listOfPoints.add(new Double(numOfLineages));

        while (numOfLineages > 1) {
            numOfLineages *= reducedPercNumOfLineagesPerInterval;
            listOfPoints.add(new Double(numOfLineages));
        }

        int lengthOfArray = listOfPoints.size();
        double[] intervalPoints = new double[ lengthOfArray ];
        for (int i = 0 ; i < lengthOfArray; i ++) {
            intervalPoints[i] = listOfPoints.get(i);
        }

        return intervalPoints;
    }

    public void initialData() {
        int nMinusOne = sampleSize - 1;

        double[] intervalPointsOfNumLineages = calculateIntervalPoints(sampleSize);
        int numOfIntervalPoints = intervalPointsOfNumLineages.length;

        SinglePopBackwardResults firstBackwardResults = new SinglePopBackwardResults();
        firstBackwardResults.time = 0;
        firstBackwardResults.startNumOfLineages = sampleSize;
        firstBackwardResults.backwardListP = new double[] {1.0d};

        SinglePopBackwardResults previousBackwardResults = firstBackwardResults;

        double oneMinusMinProbability = 1.0d - minProbability;
        maxTimeTracedBack = 0;

        double deltaT = -1;

        List<Double> listOfTimePoints = new ArrayList<Double>(numOfIntervalPoints);
        listOfTimePoints.add(new Double(0));

        List<double[]> listOfStoredBranchLength = new ArrayList<double[]>(numOfIntervalPoints);

        expectedNumOfLineages = new double[numOfIntervalPoints];
        expectedNumOfLineages[0] = sampleSize;

        double startTime = 0;
        for (int intervalIndex = 1; intervalIndex < numOfIntervalPoints; intervalIndex ++) {

            double targetNumOfLineages = intervalPointsOfNumLineages[ intervalIndex ];

            double[] k_Pk_deltaTs = new double[sampleSizePlusOne];
            double[] storedBranchLength = new double[sampleSize - 1];

            while (true) {
                deltaT = getDeltaT(previousBackwardResults, threshold_DeltaT);

                SinglePopBackwardResults currentBackwardResults = new SinglePopBackwardResults();

                double probOfMRCA = coalesce(startTime, deltaT, previousBackwardResults, currentBackwardResults);

                int startNumOfLineages = previousBackwardResults.startNumOfLineages;
                int   endNumOfLineages = startNumOfLineages + previousBackwardResults.backwardListP.length - 1;
                for (int k = startNumOfLineages; k <= endNumOfLineages; k ++) {

                    k_Pk_deltaTs[k] += k * previousBackwardResults.backwardListP[k - startNumOfLineages] * deltaT;
                }

                previousBackwardResults = currentBackwardResults;

                double numOfLineages = getExpectedNumOfLineages(currentBackwardResults);
                startTime += deltaT;

                if (numOfLineages <= targetNumOfLineages) {
                    listOfTimePoints.add(new Double(startTime));
                    expectedNumOfLineages[intervalIndex] = numOfLineages;
                    break;
                }

                if (probOfMRCA > oneMinusMinProbability) {
                    listOfTimePoints.add(new Double(startTime));
                    expectedNumOfLineages[intervalIndex] = numOfLineages;
                    maxTimeTracedBack = startTime;
                    break;
                }
            }

            for (int k = 2; k <= sampleSize; k ++) {

                int maxMarked_time0 = sampleSizePlusOne - k;
                for(int marked_time0 = 1; marked_time0 <= maxMarked_time0; marked_time0 ++) {
                    storedBranchLength[marked_time0 - 1] += k_Pk_deltaTs[k] * p_LineagesJump4LowerBound2Time0(1, k, marked_time0, sampleSize );
                }
            }

            listOfStoredBranchLength.add(storedBranchLength);
        }

        int numOfListMinusOne = listOfStoredBranchLength.size();
        int numOfList = numOfListMinusOne + 1;

        branchLengths = new double[sampleSize - 1];
        for (int i = 0; i < numOfListMinusOne; i ++) {
            double[] storedBranchLength = listOfStoredBranchLength.get(i);

            for (int size = 1; size < sampleSize; size ++) {
                branchLengths[size - 1] += storedBranchLength[size - 1];
            }
        }

        startTimeOfDetailedIntervals = new double[numOfList];
        for (int i = 0; i < numOfList; i ++) {
            startTimeOfDetailedIntervals[i] = listOfTimePoints.get(i).doubleValue();
        }


        numOfDetailedIntervals = numOfList;
        detailedMatrix = new double[numOfList][];
        for (int i = 0; i < numOfListMinusOne; i ++) {
            detailedMatrix[i] = listOfStoredBranchLength.get(i);
        }
        double[] lastDetailedMatrixLine = new double[sampleSize-1];
        for (int i = 1; i < sampleSize; i++) {
            lastDetailedMatrixLine[i-1] = 2.0d / i;
        }
        detailedMatrix[numOfListMinusOne] = lastDetailedMatrixLine;

        for (int intervalIndex = 1; intervalIndex < numOfDetailedIntervals-1; intervalIndex ++) {
            double[]    oneRow = detailedMatrix[intervalIndex];
            double[] preOneRow = detailedMatrix[intervalIndex - 1];
            for (int i = 0; i < nMinusOne; i ++) {
                oneRow[i] = oneRow[i] + preOneRow[i];
            }
        }

        System.out.println("maxTimeTracedBack = " + maxTimeTracedBack);
    }

    public double getExpectedNumOfLineages(SinglePopBackwardResults backwardResults) {
        double expectedNumOfLineages = 0;

        int size = backwardResults.backwardListP.length;
        int numOfLineages = backwardResults.startNumOfLineages;
        for(int i = 0; i < size; i ++) {
            expectedNumOfLineages += backwardResults.backwardListP[i] * numOfLineages;
            numOfLineages ++;
        }

        return expectedNumOfLineages;
    }

    protected double getDeltaT(SinglePopBackwardResults previousBackwardResults, double threshold_DeltaT) {
        int maxNumOfLineages = previousBackwardResults.startNumOfLineages + previousBackwardResults.backwardListP.length - 1;

        double deltaT = threshold_DeltaT / k_kMinusOne_dividedBy2[maxNumOfLineages];

        return deltaT;
    }

    protected void initial_k_kMinusOne_dividedBy2(int sampleSize) {
        k_kMinusOne_dividedBy2 = new double[sampleSize + 1];

        for (int k = 1; k <= sampleSize; k ++) {
            k_kMinusOne_dividedBy2[k] = (double)(k * (k - 1))  * 0.5d;
        }
    }

    protected void initial_oneOverK(int sampleSize) {
        oneOverK = new double[sampleSize + 1];

        for (int k = 1; k <= sampleSize; k ++) {
            oneOverK[k] = 1.0d / (double)k;
        }
    }

    protected double coalesce(double t, double deltaT, SinglePopBackwardResults previousBackwardResults, SinglePopBackwardResults currentBackwardResults) {

        System.arraycopy(valuesArray_empty, 0, valuesArray, 0, sampleSizePlusOne);

        int previousStartNumOfLineages = previousBackwardResults.startNumOfLineages;
        double[] previousBackwardListP = previousBackwardResults.backwardListP;
        int previousLength = previousBackwardListP.length;

        if (previousStartNumOfLineages > 1) {
            valuesArray[previousStartNumOfLineages - 1] += previousBackwardListP[0] * tranKtoL(t, previousStartNumOfLineages, previousStartNumOfLineages - 1, deltaT);
        }
        int endIndex = previousStartNumOfLineages + previousLength;
        for(int x = previousStartNumOfLineages; x < endIndex; x ++) {
            if (x < 1) {
            }
            else {
                valuesArray[x] += previousBackwardListP[x - previousStartNumOfLineages] * tranKtoL(t, x, x, deltaT);

                if( x < endIndex - 1 ) {
                    valuesArray[x] += previousBackwardListP[x + 1 - previousStartNumOfLineages] * tranKtoL(t, x+1, x, deltaT);
                }
            }
        }

        int startNumOfLineages = (valuesArray[previousStartNumOfLineages - 1] < tenMinusTwenty? previousStartNumOfLineages : previousStartNumOfLineages - 1);
        int   endNumOfLineages = (valuesArray[endIndex - 1] < tenMinusTwenty? (endIndex - 2) : (endIndex - 1) );
        int currentLength = endNumOfLineages - startNumOfLineages + 1;

        double[] currentBackwardListP = new double[currentLength];

        System.arraycopy(valuesArray, startNumOfLineages, currentBackwardListP, 0, currentLength);

        currentBackwardResults.time = t + deltaT;
        currentBackwardResults.startNumOfLineages = startNumOfLineages;
        currentBackwardResults.backwardListP = currentBackwardListP;

        return (valuesArray[1]);
    }

    protected double tranKtoL(double t, int k, int l, double deltaT){
        double p = 0;

        if ((k - l) == 1) {
            double coalProbOneUnit = k_kMinusOne_dividedBy2[k];
            p = coalProbOneUnit * deltaT;
        }
        else {
            if ( k == l ) {
                double coalProbOneUnit = k_kMinusOne_dividedBy2[k];
                p = coalProbOneUnit * deltaT;
                p  = 1 - p;
            }
        }

        p = (p <= 1.0d? p : 1.0d);
        p = (p >= 0.0d? p : 0.0d);

        return p;
    }

    private void createFileDirectory(String fileDirectory) {
        File file =new File(fileDirectory);
        if(!file .exists()  && !file .isDirectory()) {
            String str = file.getAbsolutePath();
            str = str + " not exist! Create directory!";
            System.out.println(str);
            file .mkdirs();
        }
    }

    public void saveMatrix() {
        if (loaded) {
            return;
        }

        String fileName = getTableFileName();
        try {
            createFileDirectory(tableDir);
            File outputFile = new File(tableDir, fileName);

            FileOutputStream out = new FileOutputStream(outputFile);
            String headLines = "sampleSize " + sampleSize + " numOfDetailedIntervals " + numOfDetailedIntervals + nl;
            headLines += "branchLengths ";
            for (int i = 0; i < sampleSize - 1; i ++) {
                headLines += branchLengths[i] + " ";
            }
            headLines += nl;
            out.write(headLines.getBytes());

            String nums = "NumOfLineages ";
            for (int j = 0; j < numOfDetailedIntervals; j ++) {
                nums += expectedNumOfLineages[j] + " ";
            }
            nums += nl;
            out.write(nums.getBytes());

            for (int intervalIndex = 0; intervalIndex < numOfDetailedIntervals; intervalIndex ++) {
                double[] oneRow = detailedMatrix[intervalIndex];
                String log = "t " + startTimeOfDetailedIntervals[intervalIndex] + " ";
                double sum = 0;
                for (int i = 0; i < sampleSize - 1; i ++) {
                    log += oneRow[i] + " ";
                    sum += oneRow[i];
                }
                log += nl;
                out.write(log.getBytes());
            }
            System.out.println();

            out.close();
        }
        catch (IOException e) {
            System.out.println(e.getMessage());
        }
    }

    public void setSampleSize(int n) {
        this.sampleSize = n;

        initialLogFactorials(sampleSize);

        initial_k_kMinusOne_dividedBy2(n);
        initial_oneOverK(n);

        sampleSizePlusOne = n + 1;

        valuesArray       = new double[ sampleSizePlusOne ];
        valuesArray_empty = new double[ sampleSizePlusOne ];

        expectedTreeLength = getExpectedTreeLength(n);
    }

    protected double getLogFactorial(int k) {
        return logFactorials[k];
    }

    protected void initialLogFactorials(int totalSampleSize) {
        logFactorials = new double[totalSampleSize + 1];
        logFactorials[0] = 0;

        for (int i = 1; i <= totalSampleSize; i ++) {
            logFactorials[i] = logFactorials[i-1] + Math.log(i);
        }
    }

    protected double p_LineagesJump4LowerBound2Time0(int marked_lowerBound, int k, int marked_time0, int n) {
        double logCombination1 = getLogCombination(marked_time0 - 1,marked_lowerBound - 1);
        if (logCombination1 == -1) {
            return 0;
        }
        double logCombination2 = getLogCombination(n - marked_time0 - 1, k - marked_lowerBound - 1);
        if (logCombination2 == -1) {
            return 0;
        }
        double logP = logCombination1 + logCombination2 - getLogCombination(n - 1, k - 1);

        return Math.exp(logP);
    }

    protected double getLogCombination(int n, int k) {
        if (k < 0) {
            if(n < 0) {
                return 0;
            } else {
                return -1;
            }
        }

        return (getLogFactorial(n) - getLogFactorial(n - k) - getLogFactorial(k));

    }

    public boolean isTableLoadable() {
        String fileName = getTableFileName();
        File file = new File(tableDir, fileName);

        return file.exists();
    }

    private void loadMatrix() {
        String fileName = getTableFileName();
        loaded = false;
        try {
            File file = new File(tableDir, fileName);
            FileReader fileReader = new FileReader(file);
            LineNumberReader in = new LineNumberReader(fileReader);

            String oneLine = in.readLine();
            StringTokenizer tokenizer = new StringTokenizer(oneLine);
            tokenizer.nextToken();
            int loadedSampleSize = Integer.parseInt(tokenizer.nextToken());
            if (loadedSampleSize != sampleSize) {
                System.out.println("Error! The sample size in the loaded file is different with the required sample size! ");
                System.out.println("loadedSampleSize = " + loadedSampleSize + " sampleSize = " + sampleSize);
                System.out.println("Exit...");
                System.exit(0);
            }
            tokenizer.nextToken();
            numOfDetailedIntervals = Integer.parseInt(tokenizer.nextToken());

            oneLine = in.readLine();
            tokenizer = new StringTokenizer(oneLine);
            tokenizer.nextToken();
            branchLengths = new double[sampleSize - 1];
            for (int i = 0; i < sampleSize - 1; i ++) {
                branchLengths[i] = Double.parseDouble(tokenizer.nextToken());
            }

            oneLine = in.readLine();
            tokenizer = new StringTokenizer(oneLine);
            tokenizer.nextToken();
            expectedNumOfLineages = new double[numOfDetailedIntervals];
            for (int j = 0; j < numOfDetailedIntervals; j ++) {
                expectedNumOfLineages[j] = Double.parseDouble(tokenizer.nextToken());
            }

            detailedMatrix = new double[numOfDetailedIntervals][];
            startTimeOfDetailedIntervals = new double[numOfDetailedIntervals];
            for (int i = 0; i < numOfDetailedIntervals; i ++) {
                oneLine = in.readLine();

                double[] oneRow = new double[sampleSize - 1];
                tokenizer = new StringTokenizer(oneLine);
                tokenizer.nextToken();
                startTimeOfDetailedIntervals[i] = Double.parseDouble(tokenizer.nextToken());
                for (int j = 0; j < sampleSize - 1; j ++) {
                    oneRow[j] = Double.parseDouble(tokenizer.nextToken());
                }
                detailedMatrix[i] = oneRow;
            }

            loaded = true;
            in.close();
        }
        catch (IOException e) {
            System.out.println(e.getMessage());
        }
    }

    public double[] getExpectedBranchLengths(PopSizeModelInstantChange instantChange) {
        double[] ratiosOfNiOverN0 = instantChange.ratiosOfNiOverN0;
        double[] startTimeOfIntervals = instantChange.timeIntervals;

        return getExpectedBranchLengths(ratiosOfNiOverN0, startTimeOfIntervals);
    }

    public double[] getExpectedBranchLengths(double[] ratiosOfNiOverN0, double[] startTimeOfIntervals) {
        int numOfIntervals = startTimeOfIntervals.length;

        double[] expectedBranchLengths = new double[sampleSize - 1];
        for (int i = 0; i < numOfIntervals; i ++) {
            double[] oneInterval = null;
            if (i < numOfIntervals - 1) {
                oneInterval = getExpectedBranchLengthDuringInterval(startTimeOfIntervals[i], startTimeOfIntervals[i+1]);
            }
            else {
                oneInterval = getExpectedBranchLengthDuringInterval(startTimeOfIntervals[i], Double.MAX_VALUE);
            }

            double ratio = ratiosOfNiOverN0[i];
            for (int j = 0; j < sampleSize - 1; j ++) {
                expectedBranchLengths[j] += oneInterval[j] * ratio;
            }
        }
        return expectedBranchLengths;
    }

    protected double[] getExpectedBranchLengthDuringInterval(double startTime, double endTime) {

        double[] cuvBLs_atStart = getCuvBranchLengthsGivenTime(startTime);
        double[] cuvBLs_atEnd   = getCuvBranchLengthsGivenTime(endTime);

        for (int size = 0; size < sampleSize - 1; size ++) {
            branchLengths[size] = cuvBLs_atEnd[size] - cuvBLs_atStart[size];
        }

        return branchLengths;
    }

    public double[] getCuvBranchLengthsGivenTime(double standardTimePoint) {

        int index = getIndexOfInsertValue(startTimeOfDetailedIntervals, standardTimePoint);

        double[] branchLengths = new double[sampleSize - 1];

        if (standardTimePoint == startTimeOfDetailedIntervals[index-1]) {
            if(index > 1) {
                System.arraycopy(detailedMatrix[index - 2], 0, branchLengths, 0, sampleSize-1);
            }
        }
        else if (index < numOfDetailedIntervals) {
            double  firstNum = startTimeOfDetailedIntervals[index - 1];
            double secondNum = startTimeOfDetailedIntervals[index];

            double ratio = (standardTimePoint - firstNum) / (secondNum - firstNum);

            double[] oneRowInMatrix_indexMinusOne = detailedMatrix[index - 1];
            for (int size = 1; size < sampleSize; size ++) {
                if (index == 1) {
                    branchLengths[size - 1] = oneRowInMatrix_indexMinusOne[size-1] * ratio;
                }
                else {
                    double[] oneRowInMatrix_indexMinusTwo = detailedMatrix[index - 2];

                    double  firstValue = oneRowInMatrix_indexMinusTwo[size - 1];
                    double secondValue = oneRowInMatrix_indexMinusOne[size - 1];
                    branchLengths[size - 1] = firstValue + (secondValue - firstValue) * ratio;
                }
            }
        }
        else {
            System.arraycopy(detailedMatrix[index - 1], 0, branchLengths, 0, sampleSize-1);
        }

        return branchLengths;
    }

    public int getIndexOfInsertValue(double[] values, double toBeInserted) {
        int length = values.length;
        if (values[length - 1] <= toBeInserted) {
            return length;
        }
        else {
            return getIndexOfInsertValue(values, toBeInserted, 0, values.length - 1);
        }
    }

    protected int getIndexOfInsertValue(double[] values, double toBeInserted, int low, int high) {
        int mid = (low + high) / 2;
        if (mid == low) {
            return high;
        }
        else {
            double midValue = values[mid];
            if (toBeInserted < midValue) {
                return getIndexOfInsertValue(values, toBeInserted, low, mid);
            }
            else {
                return getIndexOfInsertValue(values, toBeInserted, mid, high);
            }
        }
    }

    public String getTableFileName() {
        String fileName = "table_n" + sampleSize + ".txt";

        return fileName;
    }
    
    public int[] simulateDataAtConstantSizeModel(int sampleSize, double mutRatePerGenome, double N0) {
		int[] sfs = new int[sampleSize - 1];
		
		double[] ratiosOfNiOverN0 = {1.0d};
		double[] startTimeOfIntervals = {0.0d};
		
		double[] expectedBLs = getExpectedBranchLengths(ratiosOfNiOverN0, startTimeOfIntervals);
		
		for (int size = 1; size < sampleSize; size ++) {
			double branchLength = expectedBLs[size - 1];
			double branchLengthInGen = branchLength * 2.0d * N0;
			double expectedMut = branchLengthInGen * mutRatePerGenome;
			sfs[size - 1] = (int)(expectedMut);
			
			System.out.print(sfs[size - 1] + " ");
		}
		System.out.println();
		
		return sfs;
	}
    
    public int[] simulateDataAtBottleneckModel(int sampleSize, double mutRatePerGenome, double N0) {
		int[] sfs = new int[sampleSize - 1];
		
		double[] ratiosOfNiOverN0 = {1.0d, 0.2, 0.8};
		double[] startTimeOfIntervals = {0.0d, 0.1, 0.5};
		
		double[] expectedBLs = getExpectedBranchLengths(ratiosOfNiOverN0, startTimeOfIntervals);
		
		for (int size = 1; size < sampleSize; size ++) {
			double branchLength = expectedBLs[size - 1];
			double branchLengthInGen = branchLength * 2.0d * N0;
			double expectedMut = branchLengthInGen * mutRatePerGenome;
			sfs[size - 1] = (int)(expectedMut);
			
			System.out.print(sfs[size - 1] + " ");
		}
		System.out.println();
		
		return sfs;
	}
    
    /**
     * time is the stardard coalescent time
     */
    public double getAvgNumOfLineages(double time) {
    	double avgNumOfLineages = expectedNumOfLineages[ numOfDetailedIntervals - 1 ];
    	
    	if (startTimeOfDetailedIntervals[numOfDetailedIntervals - 1] <= time) {
    		double startTime = startTimeOfDetailedIntervals[numOfDetailedIntervals - 2];
    		double   endTime = startTimeOfDetailedIntervals[numOfDetailedIntervals - 1];
    		
    		double startLineages = expectedNumOfLineages[ numOfDetailedIntervals - 2 ];
    		double   endLineages = expectedNumOfLineages[ numOfDetailedIntervals - 1 ];
    		
    		double downRatio = (startLineages - endLineages) / (endTime - startTime);
    		
    		avgNumOfLineages = endLineages - (time - endTime) * downRatio;
    		
    		if (avgNumOfLineages < 1.0d) avgNumOfLineages = 1.0d;
    	}
    	else {
    		for (int i = 0; i < numOfDetailedIntervals - 1; i ++) {
        		double startTime = startTimeOfDetailedIntervals[i];
        		double   endTime = startTimeOfDetailedIntervals[i + 1];
        		
        		if (startTime <= time && time < endTime) {
        			double ratio = (time - startTime) / (endTime - startTime);
        			
        			double startLineages = expectedNumOfLineages[i];
        			double   endLineages = expectedNumOfLineages[i + 1];
        			
        			avgNumOfLineages = startLineages - (startLineages - endLineages) * ratio;
        			
        			break;
        		}
        	}
    	}
    	
    	return avgNumOfLineages;
    }
    
}
