import org.kohsuke.args4j.CmdLineParser;

import java.io.*;


public class TruncateSFS {
    public static final int NUM_STD = 3;

    private int windowLength (int[] sfs) {
    	
        int l = sfs.length;
        if (l < 50) {
            return 3;
        } else if (l < 100) {
            return 4;
        } else if (l < 300) {
            return 5;
        } else if (l < 500) {
            return 10;
        } else if (l < 1000) {
            return 20;
        } else if (l < 2000) {
            return 30;
        } else if (l < 3000) {
            return 40;
        } else{
            return 50; // l >= 3000, or n > 3000
        }
    }

    private double[] averageOfWindow (int[] sfs) {
        int wl = windowLength(sfs);
        double[] average = new double[sfs.length - wl + 1];
        for (int i = 0; i < average.length; i++) {
            average[i] = 0.0;
            for (int j = 0; j < wl; j++) {
                average[i] += sfs[i + j];
            }
            average[i] /= (double) wl;
        }
        return average;
    }

    private int[] minInWindows (int[] sfs) {
        int wl = windowLength(sfs);
        int[] mins = new int[sfs.length - wl + 1];
        for (int i = 0; i < mins.length; i++) {
            mins[i] = i;
            int m = sfs[i];
            for (int j = 1; j < wl; j++) {
                if (sfs[i + j] < m) {
                    mins[i] = i + j;
                    m = sfs[i + j];
                }
            }
        }
        return mins;
    }

    private double[] standardDeviation (double[] average) {
        double[] std = new double[average.length];
        for (int i = 0; i < std.length; i++) {
            std[i] = Math.sqrt(average[i]);
        }
        return std;
    }

    private double[] upperBound (double[] average, double[] std) {
        double[] upper = new double[average.length];
        for (int i = 0; i < upper.length; i++) {
            upper[i] = average[i] + std[i] * NUM_STD;
        }
        return upper;
    }

    private double[] lowerBound (double[] average, double[] std) {
        double[] lower = new double[average.length];
        for (int i = 0; i < lower.length; i++) {
            lower[i] = average[i] - std[i] * NUM_STD;
        }
        return lower;
    }

    private int[] countSitesWithinBound (int[] sfs, double[] upperBound, double[] lowerBound) {
        int wl = windowLength(sfs);
        int[] countSites = new int[upperBound.length];
        for (int i = upperBound.length - 1; i >= 0; i--) {
            countSites[i] = 0;
            for (int j = 0; j < wl; j++) {
                if ((sfs[i + j] >= lowerBound[i]) && (sfs[i + j] <= upperBound[i])) {
                    countSites[i]++;
                }
            }
        }
        return countSites;
    }

    private void cutSFS (int[] sfs) {
        int omitNum = 0;
        double[] average = averageOfWindow(sfs);
        double[] std = standardDeviation(average);
        double[] upper = upperBound(average, std);
        double[] lower = lowerBound(average, std);
        int[] countSites = countSitesWithinBound(sfs, upper, lower);
        int[] mins = minInWindows(sfs);
        int wl = windowLength(sfs);

        for (int i = countSites.length - 1; i >= 2 ; i--) {
            if (countSites[i] == wl && average[i] <= average[i - 1]) {
                omitNum = countSites.length - i - 1;
                
                System.out.println("To avoid the effects of positive selection and population structure, the lifted tail of SFS can be removed or collapsed.");
                System.out.println();
                System.out.println("The number of SFS types to be truncated or collapsed: " + omitNum + " (recommended)");
                System.out.println("The proportion of truncated or collapsed SFS types: " + (double) omitNum / (double)sfs.length);
                System.out.println();
                
                break;
            }
        }
    }

    private void run (String inputFile) throws IOException {

        FileReader fileReader = new FileReader(inputFile);
        LineNumberReader lineNumberReader = new LineNumberReader(fileReader);
        String line;
        int dataIndex = 1;
        while ((line = lineNumberReader.readLine()) != null) {
            String[] data = line.trim().split("\\s+");
            int[] sfs = new int[data.length];
            for (int i = 0; i < sfs.length; i++) {
                sfs[i] = Integer.parseInt(data[i]);
            }
            System.out.println("Data: " + dataIndex++);
            
            int n = sfs.length + 1;
        	System.out.println("n = " + n);
        	
            cutSFS(sfs);
        }
    }

    public static void main(String[] args) throws Exception {
        TruncateSFSInput truncateSFSInput = new TruncateSFSInput();
        CmdLineParser parser = new CmdLineParser(truncateSFSInput);
        parser.parseArgument(args);

        TruncateSFS truncateSFS = new TruncateSFS();
        String inputFile = truncateSFSInput.getInputFile();
        truncateSFS.run(inputFile);
    }
}
